# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Monte-Carlo variants of the LogEI family of improvements-based acquisition functions,
see [Ament2023logei]_ for details.

References

.. [Ament2023logei]
    S. Ament, S. Daulton, D. Eriksson, M. Balandat, and E. Bakshy.
    Unexpected Improvements to Expected Improvement for Bayesian Optimization. Advances
    in Neural Information Processing Systems 36, 2023.
"""

from __future__ import annotations

from collections.abc import Callable
from copy import deepcopy
from functools import partial
from typing import TypeVar

import torch
from botorch.acquisition.cached_cholesky import CachedCholeskyMCSamplerMixin
from botorch.acquisition.monte_carlo import SampleReducingMCAcquisitionFunction
from botorch.acquisition.objective import (
    ConstrainedMCObjective,
    IdentityMCObjective,
    MCAcquisitionObjective,
    PosteriorTransform,
)
from botorch.acquisition.utils import (
    compute_best_feasible_objective,
    prune_inferior_points,
)
from botorch.exceptions.errors import BotorchError
from botorch.models.model import Model
from botorch.sampling.base import MCSampler
from botorch.utils.objective import compute_smoothed_feasibility_indicator
from botorch.utils.safe_math import (
    fatmax,
    log_fatplus,
    log_softplus,
    logmeanexp,
    smooth_amax,
)
from botorch.utils.transforms import match_batch_shape
from torch import Tensor

"""
NOTE: On the default temperature parameters:

tau_relu: It is generally important to set `tau_relu` to be very small, in particular,
smaller than the expected improvement value. Otherwise, the optimization can stagnate.
By setting `tau_relu=1e-6` by default, stagnation is exceedingly unlikely to occur due
to the smooth ReLU approximation for practical applications of BO.
IDEA: We could consider shrinking `tau_relu` with the progression of the optimization.

tau_max: This is only relevant for the batch (`q > 1`) case, and `tau_max=1e-2` is
sufficient to get a good approximation to the maximum improvement in the batch of
candidates. If `fat=False`, the smooth approximation to the maximum can saturate
numerically. It is therefore recommended to use `fat=True` when optimizing batches
of `q > 1` points.
"""
TAU_RELU = 1e-6
TAU_MAX = 1e-2
FloatOrTensor = TypeVar("FloatOrTensor", float, Tensor)


class LogImprovementMCAcquisitionFunction(SampleReducingMCAcquisitionFunction):
    r"""
    Abstract base class for Monte-Carlo-based batch LogEI acquisition functions.
    """

    _log: bool = True

    def __init__(
        self,
        model: Model,
        sampler: MCSampler | None = None,
        objective: MCAcquisitionObjective | None = None,
        posterior_transform: PosteriorTransform | None = None,
        X_pending: Tensor | None = None,
        constraints: list[Callable[[Tensor], Tensor]] | None = None,
        eta: Tensor | float = 1e-3,
        fat: bool = True,
        tau_max: float = TAU_MAX,
    ) -> None:
        r"""Constructor of the base class for LogEI acquisition functions.

        Args:
            model: A fitted model.
            sampler: The sampler used to draw base samples. If not given,
                a sampler is generated using `get_sampler`.
                NOTE: For posteriors that do not support base samples,
                a sampler compatible with intended use case must be provided.
                See `ForkedRNGSampler` and `StochasticSampler` as examples.
            objective: The MCAcquisitionObjective under which the samples are
                evaluated. Defaults to `IdentityMCObjective()`.
            posterior_transform: A PosteriorTransform (optional).
            X_pending: A `batch_shape, m x d`-dim Tensor of `m` design points
                that have points that have been submitted for function evaluation
                but have not yet been evaluated.
            constraints: A list of constraint callables which map a Tensor of posterior
                samples of dimension `sample_shape x batch-shape x q x m`-dim to a
                `sample_shape x batch-shape x q`-dim Tensor. The associated constraints
                are satisfied if `constraint(samples) < 0`.
            eta: Temperature parameter(s) governing the smoothness of the sigmoid
                approximation to the constraint indicators. See the docs of
                `compute_(log_)constraint_indicator` for more details on this parameter.
            fat: Toggles the logarithmic / linear asymptotic behavior of the smooth
                approximation to the ReLU.
            tau_max: Temperature parameter controlling the sharpness of the
                approximation to the `max` operator over the `q` candidate points.
        """
        if isinstance(objective, ConstrainedMCObjective):
            raise BotorchError(
                "Log-Improvement should not be used with `ConstrainedMCObjective`."
                "Please pass the `constraints` directly to the constructor of the "
                "acquisition function."
            )
        q_reduction = partial(fatmax if fat else smooth_amax, tau=tau_max)
        super().__init__(
            model=model,
            sampler=sampler,
            objective=objective,
            posterior_transform=posterior_transform,
            X_pending=X_pending,
            sample_reduction=logmeanexp,
            q_reduction=q_reduction,
            constraints=constraints,
            eta=eta,
            fat=fat,
        )
        self.tau_max = tau_max


class qLogExpectedImprovement(LogImprovementMCAcquisitionFunction):
    r"""MC-based batch Log Expected Improvement.

    This computes qLogEI by
    (1) sampling the joint posterior over q points,
    (2) evaluating the smoothed log improvement over the current best for each sample,
    (3) smoothly maximizing over q, and
    (4) averaging over the samples in log space.

    See [Ament2023logei]_ for details. Formally,

    `qLogEI(X) ~ log(qEI(X)) = log(E(max(max Y - best_f, 0)))`.

    where `Y ~ f(X)`, and `X = (x_1,...,x_q)`, .

    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> best_f = train_Y.max()[0]
        >>> sampler = SobolQMCNormalSampler(1024)
        >>> qLogEI = qLogExpectedImprovement(model, best_f, sampler)
        >>> qei = qLogEI(test_X)
    """

    def __init__(
        self,
        model: Model,
        best_f: float | Tensor,
        sampler: MCSampler | None = None,
        objective: MCAcquisitionObjective | None = None,
        posterior_transform: PosteriorTransform | None = None,
        X_pending: Tensor | None = None,
        constraints: list[Callable[[Tensor], Tensor]] | None = None,
        eta: Tensor | float = 1e-3,
        fat: bool = True,
        tau_max: float = TAU_MAX,
        tau_relu: float = TAU_RELU,
    ) -> None:
        r"""q-Log Expected Improvement.

        Args:
            model: A fitted model.
            best_f: The best objective value observed so far (assumed noiseless). Can be
                a scalar, or a `batch_shape`-dim tensor. In case of a batched model, the
                tensor can specify different values for each element of the batch.
            sampler: The sampler used to draw base samples. See `MCAcquisitionFunction`
                more details.
            objective: The MCAcquisitionObjective under which the samples are evaluated.
                Defaults to `IdentityMCObjective()`.
            posterior_transform: A PosteriorTransform (optional).
            X_pending:  A `m x d`-dim Tensor of `m` design points that have been
                submitted for function evaluation but have not yet been evaluated.
                Concatenated into `X` upon forward call. Copied and set to have no
                gradient.
            constraints: A list of constraint callables which map a Tensor of posterior
                samples of dimension `sample_shape x batch-shape x q x m`-dim to a
                `sample_shape x batch-shape x q`-dim Tensor. The associated constraints
                are satisfied if `constraint(samples) < 0`.
            eta: Temperature parameter(s) governing the smoothness of the sigmoid
                approximation to the constraint indicators. See the docs of
                `compute_(log_)smoothed_constraint_indicator` for details.
            fat: Toggles the logarithmic / linear asymptotic behavior of the smooth
                approximation to the ReLU.
            tau_max: Temperature parameter controlling the sharpness of the smooth
                approximations to max.
            tau_relu: Temperature parameter controlling the sharpness of the smooth
                approximations to ReLU.
        """
        super().__init__(
            model=model,
            sampler=sampler,
            objective=objective,
            posterior_transform=posterior_transform,
            X_pending=X_pending,
            constraints=constraints,
            eta=eta,
            tau_max=check_tau(tau_max, name="tau_max"),
            fat=fat,
        )
        self.register_buffer("best_f", torch.as_tensor(best_f))
        self.tau_relu = check_tau(tau_relu, name="tau_relu")

    def _sample_forward(self, obj: Tensor) -> Tensor:
        r"""Evaluate qLogExpectedImprovement on the candidate set `X`.

        Args:
            obj: `mc_shape x batch_shape x q`-dim Tensor of MC objective values.

        Returns:
            A `mc_shape x batch_shape x q`-dim Tensor of expected improvement values.
        """
        li = _log_improvement(
            Y=obj,
            best_f=self.best_f,
            tau=self.tau_relu,
            fat=self._fat,
        )
        return li


class qLogNoisyExpectedImprovement(
    LogImprovementMCAcquisitionFunction, CachedCholeskyMCSamplerMixin
):
    r"""MC-based batch Log Noisy Expected Improvement.

    This function does not assume a `best_f` is known (which would require
    noiseless observations). Instead, it uses samples from the joint posterior
    over the `q` test points and previously observed points. A smooth approximation
    to the canonical improvement over previously observed points is computed
    for each sample and the logarithm of the average is returned.

    See [Ament2023logei]_ for details. Formally,

    `qLogNEI(X) ~ log(qNEI(X)) = Log E(max(max Y - max Y_baseline, 0))`,

    where `(Y, Y_baseline) ~ f((X, X_baseline)), X = (x_1,...,x_q)`.

    For optimizing a batch of `q > 1` points using sequential greedy optimization,
    the incremental improvement from the latest point is computed and returned by
    default. I.e. the pending points are treated X_baseline. Often, the incremental
    EI is easier to optimize.

    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> sampler = SobolQMCNormalSampler(1024)
        >>> qLogNEI = qLogNoisyExpectedImprovement(model, train_X, sampler)
        >>> acqval = qLogNEI(test_X)
    """

    def __init__(
        self,
        model: Model,
        X_baseline: Tensor,
        sampler: MCSampler | None = None,
        objective: MCAcquisitionObjective | None = None,
        posterior_transform: PosteriorTransform | None = None,
        X_pending: Tensor | None = None,
        constraints: list[Callable[[Tensor], Tensor]] | None = None,
        eta: Tensor | float = 1e-3,
        fat: bool = True,
        prune_baseline: bool = True,
        cache_root: bool = True,
        tau_max: float = TAU_MAX,
        tau_relu: float = TAU_RELU,
        marginalize_dim: int | None = None,
        incremental: bool = True,
    ) -> None:
        r"""q-Noisy Expected Improvement.

        Args:
            model: A fitted model.
            X_baseline: A `batch_shape x r x d`-dim Tensor of `r` design points
                that have already been observed. These points are considered as
                the potential best design point.
            sampler: The sampler used to draw base samples. See `MCAcquisitionFunction`
                more details.
            objective: The MCAcquisitionObjective under which the samples are
                evaluated. Defaults to `IdentityMCObjective()`.
            posterior_transform: A PosteriorTransform (optional).
            X_pending: A `batch_shape x m x d`-dim Tensor of `m` design points
                that have points that have been submitted for function evaluation
                but have not yet been evaluated. Concatenated into `X` upon
                forward call. Copied and set to have no gradient.
            constraints: A list of constraint callables which map a Tensor of posterior
                samples of dimension `sample_shape x batch-shape x q x m`-dim to a
                `sample_shape x batch-shape x q`-dim Tensor. The associated constraints
                are satisfied if `constraint(samples) < 0`.
            eta: Temperature parameter(s) governing the smoothness of the sigmoid
                approximation to the constraint indicators. See the docs of
                `compute_(log_)smoothed_constraint_indicator` for details.
            fat: Toggles the logarithmic / linear asymptotic behavior of the smooth
                approximation to the ReLU.
            prune_baseline: If True, remove points in `X_baseline` that are
                highly unlikely to be the best point. This can significantly
                improve performance and is generally recommended. In order to
                customize pruning parameters, instead manually call
                `botorch.acquisition.utils.prune_inferior_points` on `X_baseline`
                before instantiating the acquisition function.
            cache_root: A boolean indicating whether to cache the root
                decomposition over `X_baseline` and use low-rank updates.
            tau_max: Temperature parameter controlling the sharpness of the smooth
                approximations to max.
            tau_relu: Temperature parameter controlling the sharpness of the smooth
                approximations to ReLU.
            marginalize_dim: The dimension to marginalize over.
            incremental: Whether to compute incremental EI over the pending points
                or compute EI of the joint batch improvement (including pending
                points).

        TODO: similar to qNEHVI, when we are using sequential greedy candidate
        selection, we could incorporate pending points X_baseline and compute
        the incremental q(Log)NEI from the new point. This would greatly increase
        efficiency for large batches.
        """
        # TODO: separate out baseline variables initialization and other functions
        # in qNEI to avoid duplication of both code and work at runtime.
        self.incremental = incremental

        super().__init__(
            model=model,
            sampler=sampler,
            objective=objective,
            posterior_transform=posterior_transform,
            # we set X_pending in init_baseline for incremental NEI
            X_pending=X_pending if not incremental else None,
            constraints=constraints,
            eta=eta,
            fat=fat,
            tau_max=tau_max,
        )
        self.tau_relu = tau_relu
        self.prune_baseline = prune_baseline
        self.marginalize_dim = marginalize_dim
        if incremental:
            self.X_pending = None  # required to initialize attribute for optimize_acqf
        self._init_baseline(
            model=model,
            X_baseline=X_baseline,
            # This is ignored in incremental=False
            X_pending=X_pending,
            sampler=sampler,
            objective=objective,
            posterior_transform=posterior_transform,
            cache_root=cache_root,
        )

    def _sample_forward(self, obj: Tensor) -> Tensor:
        r"""Evaluate qLogNoisyExpectedImprovement per sample on the candidate set `X`.

        Args:
            obj: `mc_shape x batch_shape x q`-dim Tensor of MC objective values.

        Returns:
            A `sample_shape x batch_shape x q`-dim Tensor of log noisy expected smoothed
            improvement values.
        """
        return _log_improvement(
            Y=obj,
            best_f=self.compute_best_f(obj),
            tau=self.tau_relu,
            fat=self._fat,
        )

    def _init_baseline(
        self,
        model: Model,
        X_baseline: Tensor,
        X_pending: Tensor | None = None,
        sampler: MCSampler | None = None,
        objective: MCAcquisitionObjective | None = None,
        posterior_transform: PosteriorTransform | None = None,
        cache_root: bool = True,
    ) -> None:
        CachedCholeskyMCSamplerMixin.__init__(
            self, model=model, cache_root=cache_root, sampler=sampler
        )
        if self.prune_baseline:
            X_baseline = prune_inferior_points(
                model=model,
                X=X_baseline,
                objective=objective,
                posterior_transform=posterior_transform,
                marginalize_dim=self.marginalize_dim,
                constraints=self._constraints,
            )
        self.register_buffer("_X_baseline", X_baseline)
        # full_X_baseline is the set of points that should be considered as the
        # incumbent. For incremental EI, this contains the previously evaluated
        # points (X_baseline) and pending points (X_pending). For non-incremental
        # EI, this contains the  previously evaluated points (X_baseline).
        if X_pending is not None and self.incremental:
            full_X_baseline = torch.cat([X_baseline, X_pending], dim=-2)
        else:
            full_X_baseline = X_baseline
        self.register_buffer("_full_X_baseline", full_X_baseline)
        # registering buffers for _get_samples_and_objectives in the next `if` block
        self.register_buffer("baseline_samples", None)
        self.register_buffer("baseline_obj", None)
        if self._cache_root:
            self.q_in = -1
            # set baseline samples
            with torch.no_grad():  # this is _get_samples_and_objectives(X_baseline)
                posterior = self.model.posterior(
                    self.X_baseline, posterior_transform=self.posterior_transform
                )
                # Note: The root decomposition is cached in two different places. It
                # may be confusing to have two different caches, but this is not
                # trivial to change since each is needed for a different reason:
                # - LinearOperator caching to `posterior.mvn` allows for reuse within
                #   this function, which may be helpful if the same root decomposition
                #   is produced by the calls to `self.base_sampler` and
                #   `self._cache_root_decomposition`.
                # - self._baseline_L allows a root decomposition to be persisted outside
                #   this method.
                self.baseline_samples = self.get_posterior_samples(posterior)
                self.baseline_obj = self.objective(
                    self.baseline_samples, X=self.X_baseline
                )

            # We make a copy here because we will write an attribute `base_samples`
            # to `self.base_sampler.base_samples`, and we don't want to mutate
            # `self.sampler`.
            self.base_sampler = deepcopy(self.sampler)
            self.register_buffer(
                "_baseline_best_f",
                self._compute_best_feasible_objective(
                    samples=self.baseline_samples, obj=self.baseline_obj
                ),
            )
            self._baseline_L = self._compute_root_decomposition(posterior=posterior)

    @property
    def X_baseline(self) -> Tensor:
        """Returns the set of points that should be considered as the incumbent.

        For incremental EI, this contains the previously evaluated points
        (X_baseline) and pending points (X_pending). For non-incremental
        EI, this contains the  previously evaluated points (X_baseline).
        """
        return self._full_X_baseline

    def set_X_pending(self, X_pending: Tensor | None = None) -> None:
        r"""Informs the acquisition function about pending design points.

        Here pending points are concatenated with X_baseline and incremental
        NEI is computed.

        Args:
            X_pending: `n x d` Tensor with `n` `d`-dim design points that have
                been submitted for evaluation but have not yet been evaluated.
        """
        if not self.incremental:
            return super().set_X_pending(X_pending=X_pending)
        if X_pending is None:
            if not hasattr(self, "_full_X_baseline") or (
                self._full_X_baseline.shape[-2] == self._X_baseline.shape[-2]
            ):
                return
            else:
                # reset pending points
                X_pending = None
        self._init_baseline(
            model=self.model,
            X_baseline=self._X_baseline,
            X_pending=X_pending,
            sampler=self.sampler,
            objective=self.objective,
            posterior_transform=self.posterior_transform,
            cache_root=self._cache_root,
        )

    def compute_best_f(self, obj: Tensor) -> Tensor:
        """Computes the best (feasible) noisy objective value.

        Args:
            obj: `sample_shape x batch_shape x q`-dim Tensor of objectives in forward.

        Returns:
            A `sample_shape x batch_shape`-dim Tensor of best feasible objectives.
        """
        if self._cache_root:
            val = self._baseline_best_f
        else:
            val = self._compute_best_feasible_objective(
                samples=self.baseline_samples, obj=self.baseline_obj
            )
        # ensuring shape, dtype, device compatibility with obj
        n_sample_dims = len(self.sample_shape)
        view_shape = torch.Size(
            [
                *val.shape[:n_sample_dims],  # sample dimensions
                *(1,) * (obj.ndim - val.ndim - 1),  # pad to match obj without `q`-dim
                *val.shape[n_sample_dims:],  # the rest
            ]
        )
        return val.view(view_shape).to(obj)  # obj.shape[:-1], i.e. without `q`-dim`

    def _get_samples_and_objectives(self, X: Tensor) -> tuple[Tensor, Tensor]:
        r"""Compute samples at new points, using the cached root decomposition.

        Args:
            X: A `batch_shape x q x d`-dim tensor of inputs.

        Returns:
            A two-tuple `(samples, obj)`, where `samples` is a tensor of posterior
            samples with shape `sample_shape x batch_shape x q x m`, and `obj` is a
            tensor of MC objective values with shape `sample_shape x batch_shape x q`.
        """
        n_baseline, q = self.X_baseline.shape[-2], X.shape[-2]
        X_full = torch.cat([match_batch_shape(self.X_baseline, X), X], dim=-2)
        # TODO: Implement more efficient way to compute posterior over both training and
        # test points in GPyTorch (https://github.com/cornellius-gp/gpytorch/issues/567)
        posterior = self.model.posterior(
            X_full, posterior_transform=self.posterior_transform
        )
        if not self._cache_root:
            samples_full = super().get_posterior_samples(posterior)
            obj_full = self.objective(samples_full, X=X_full)
            # Calculate the positive index for splitting the samples & objective values.
            split_dim = len(obj_full.shape) - 1
            # assigning baseline buffers so `best_f` can be computed in _sample_forward
            self.baseline_samples, samples = samples_full.split(
                [n_baseline, q], dim=split_dim
            )
            self.baseline_obj, obj = obj_full.split([n_baseline, q], dim=split_dim)
            return samples, obj

        # handle one-to-many input transforms
        n_plus_q = X_full.shape[-2]
        n_w = posterior._extended_shape()[-2] // n_plus_q
        q_in = q * n_w
        self._set_sampler(q_in=q_in, posterior=posterior)
        samples = self._get_f_X_samples(posterior=posterior, q_in=q_in)
        obj = self.objective(samples, X=X_full[..., -q:, :])
        return samples, obj

    def _compute_best_feasible_objective(self, samples: Tensor, obj: Tensor) -> Tensor:
        r"""Computes best feasible objective value from samples.

        Args:
            samples: `sample_shape x batch_shape x q x m`-dim posterior samples.
            obj: A `sample_shape x batch_shape x q`-dim Tensor of MC objective values.

        Returns:
            A `sample_shape x batch_shape`-dim Tensor of best feasible objectives.
        """
        return compute_best_feasible_objective(
            samples=samples,
            obj=obj,
            constraints=self._constraints,
            model=self.model,
            objective=self.objective,
            posterior_transform=self.posterior_transform,
            X_baseline=self.X_baseline,
        )


class qLogProbabilityOfFeasibility(LogImprovementMCAcquisitionFunction):
    r"""MC-based batch LogProbabilityOfFeasibility.

    This computes the log probability of feasibility by
    (1) sampling the joint posterior over q points
    (2) evaluating the feasibility of each sample
    (3) averaging over the sample and batch dimensions.

    `log_prob_feas(X) = log(P(f(X) <= 0)), where f(X) ~ GP`.
    """

    def __init__(
        self,
        model: Model,
        constraints: list[Callable[[Tensor], Tensor]],
        sampler: MCSampler | None = None,
        objective: MCAcquisitionObjective | None = None,
        posterior_transform: PosteriorTransform | None = None,
        X_pending: Tensor | None = None,
        eta: Tensor | float = 1e-3,
        fat: bool = True,
        tau_max: float = TAU_MAX,
    ):
        r"""Constructor of the batch log probability of feasibility acquisition
        function.

        Args:
            model: A fitted model.
            constraints: A list of constraint callables which map a Tensor of posterior
                samples of dimension `sample_shape x batch-shape x q x m`-dim to a
                `sample_shape x batch-shape x q`-dim Tensor. The associated constraints
                are satisfied if `constraint(samples) < 0`.
            sampler: The sampler used to draw base samples. If not given,
                a sampler is generated using `get_sampler`.
                NOTE: For posteriors that do not support base samples,
                a sampler compatible with intended use case must be provided.
                See `ForkedRNGSampler` and `StochasticSampler` as examples.
            objective: Not used, kept for compatibility with interface.
            posterior_transform: A PosteriorTransform (optional).
            X_pending: A `batch_shape, m x d`-dim Tensor of `m` design points
                that have points that have been submitted for function evaluation
                but have not yet been evaluated.
            eta: Temperature parameter(s) governing the smoothness of the sigmoid
                approximation to the constraint indicators. See the docs of
                `compute_(log_)constraint_indicator` for more details on this parameter.
            fat: Toggles the logarithmic / linear asymptotic behavior of the smooth
                approximation to the ReLU.
            tau_max: Temperature parameter controlling the sharpness of the
                approximation to the `max` operator over the `q` candidate points.
        """
        super().__init__(
            model=model,
            sampler=sampler,
            # theoretically we don't need an objective, but MCAcquisitionFunction will
            # throw an error if the model has multiple outputs and the objective is None
            objective=IdentityMCObjective(),
            posterior_transform=posterior_transform,
            X_pending=X_pending,
            constraints=constraints,
            eta=eta,
            fat=fat,
            tau_max=tau_max,
        )

    def _non_reduced_forward(self, X: Tensor) -> Tensor:
        """Compute the feasibility values at the MC-sample and batch (q) level.

        Args:
            X: A `batch_shape x q x d` Tensor of t-batches with `q` `d`-dim
                design points each.

        Returns:
            A Tensor with shape `sample_sample x batch_shape x q`.
        """
        posterior = self.model.posterior(
            X=X, posterior_transform=self.posterior_transform
        )
        return compute_smoothed_feasibility_indicator(
            constraints=self._constraints,
            samples=self.get_posterior_samples(posterior),
            eta=self._eta,
            log=self._log,
            fat=self._fat,
        )

    def _sample_forward(self, obj: Tensor) -> Tensor:
        """Not necessary, since we are implementing `_non_reduced_forward` without it.
        Need to provide an implementation to avoid an abstract method error.
        """
        raise NotImplementedError("This should be dead code.")  # pragma: no cover


"""
###################################### utils ##########################################
"""


def _log_improvement(
    Y: Tensor,
    best_f: Tensor,
    tau: float | Tensor,
    fat: bool,
) -> Tensor:
    """Computes the logarithm of the softplus-smoothed improvement, i.e.
    `log_softplus(Y - best_f, beta=(1 / tau))`.
    Note that softplus is an approximation to the regular ReLU objective whose maximum
    pointwise approximation error is linear with respect to tau as tau goes to zero.

    Args:
        obj: `mc_samples x batch_shape x q`-dim Tensor of output samples.
        best_f: Best previously observed objective value(s), broadcastable with
            `mc_samples x batch_shape`-dim Tensor, i.e. `obj`'s dims without `q`.
        tau: Temperature parameter for smooth approximation of ReLU.
            as `tau -> 0`, maximum pointwise approximation error is linear w.r.t. `tau`.
        fat: Toggles the logarithmic / linear asymptotic behavior of the
            smooth approximation to ReLU.

    Returns:
        A `mc_samples x batch_shape x q`-dim Tensor of improvement values.
    """
    log_soft_clamp = log_fatplus if fat else log_softplus
    Z = Y - best_f.unsqueeze(-1).to(Y)
    return log_soft_clamp(Z, tau=tau)  # ~ ((Y - best_f) / Y_std).clamp(0)


def check_tau(tau: FloatOrTensor, name: str) -> FloatOrTensor:
    """Checks the validity of the tau arguments of the functions below, and returns
    `tau` if it is valid."""
    if isinstance(tau, Tensor) and tau.numel() != 1:
        raise ValueError(f"{name} is not a scalar: {tau.numel()=}.")
    if not (tau > 0):
        raise ValueError(f"{name} is non-positive: {tau=}.")
    return tau



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable
from typing import Any, overload, Union

import torch
from botorch.models.approximate_gp import SingleTaskVariationalGP
from botorch.models.gpytorch import GPyTorchModel
from botorch.models.model import Model, ModelList
from botorch.models.transforms.input import InputTransform
from botorch.models.transforms.outcome import OutcomeTransform
from botorch.utils.dispatcher import Dispatcher
from gpytorch.kernels import ScaleKernel
from gpytorch.kernels.kernel import Kernel
from torch import LongTensor, Tensor
from torch.nn import Module, ModuleList

TInputTransform = Union[InputTransform, Callable[[Tensor], Tensor]]
TOutputTransform = Union[OutcomeTransform, Callable[[Tensor], Tensor]]
GetTrainInputs = Dispatcher("get_train_inputs")
GetTrainTargets = Dispatcher("get_train_targets")


class TransformedModuleMixin:
    r"""Mixin that wraps a module's __call__ method with optional transforms."""

    input_transform: TInputTransform | None
    output_transform: TOutputTransform | None

    def __call__(self, values: Tensor, *args: Any, **kwargs: Any) -> Tensor:
        input_transform = getattr(self, "input_transform", None)
        if input_transform is not None:
            values = (
                input_transform.forward(values)
                if isinstance(input_transform, InputTransform)
                else input_transform(values)
            )

        output = super().__call__(values, *args, **kwargs)
        output_transform = getattr(self, "output_transform", None)
        if output_transform is None:
            return output

        return (
            output_transform.untransform(output)[0]
            if isinstance(output_transform, OutcomeTransform)
            else output_transform(output)
        )


class TensorTransform(ABC, Module):
    r"""Abstract base class for transforms that map tensor to tensor."""

    @abstractmethod
    def forward(self, values: Tensor, **kwargs: Any) -> Tensor:
        pass  # pragma: no cover


class ChainedTransform(TensorTransform):
    r"""A composition of TensorTransforms."""

    def __init__(self, *transforms: TensorTransform):
        r"""Initializes a ChainedTransform instance.

        Args:
            transforms: A set of transforms to be applied from right to left.
        """
        super().__init__()
        self.transforms = ModuleList(transforms)

    def forward(self, values: Tensor) -> Tensor:
        for transform in reversed(self.transforms):
            values = transform(values)
        return values


class SineCosineTransform(TensorTransform):
    r"""A transform that returns concatenated sine and cosine features."""

    def __init__(self, scale: Tensor | None = None):
        r"""Initializes a SineCosineTransform instance.

        Args:
            scale: An optional tensor used to rescale the module's outputs.
        """
        super().__init__()
        self.scale = scale

    def forward(self, values: Tensor) -> Tensor:
        sincos = torch.concat([values.sin(), values.cos()], dim=-1)
        return sincos if self.scale is None else self.scale * sincos


class InverseLengthscaleTransform(TensorTransform):
    r"""A transform that divides its inputs by a kernels lengthscales."""

    def __init__(self, kernel: Kernel):
        r"""Initializes an InverseLengthscaleTransform instance.

        Args:
            kernel: The kernel whose lengthscales are to be used.
        """
        if not kernel.has_lengthscale:
            raise RuntimeError(f"{type(kernel)} does not implement `lengthscale`.")

        super().__init__()
        self.kernel = kernel

    def forward(self, values: Tensor) -> Tensor:
        return self.kernel.lengthscale.reciprocal() * values


class OutputscaleTransform(TensorTransform):
    r"""A transform that multiplies its inputs by the square root of a
    kernel's outputscale."""

    def __init__(self, kernel: ScaleKernel):
        r"""Initializes an OutputscaleTransform instance.

        Args:
            kernel: A ScaleKernel whose `outputscale` is to be used.
        """
        super().__init__()
        self.kernel = kernel

    def forward(self, values: Tensor) -> Tensor:
        outputscale = (
            self.kernel.outputscale[..., None, None]
            if self.kernel.batch_shape
            else self.kernel.outputscale
        )
        return outputscale.sqrt() * values


class FeatureSelector(TensorTransform):
    r"""A transform that returns a subset of its input's features.
    along a given tensor dimension."""

    def __init__(self, indices: Iterable[int], dim: int | LongTensor = -1):
        r"""Initializes a FeatureSelector instance.

        Args:
            indices: A LongTensor of feature indices.
            dim: The dimensional along which to index features.
        """
        super().__init__()
        self.register_buffer("dim", dim if torch.is_tensor(dim) else torch.tensor(dim))
        self.register_buffer(
            "indices", indices if torch.is_tensor(indices) else torch.tensor(indices)
        )

    def forward(self, values: Tensor) -> Tensor:
        return values.index_select(dim=self.dim, index=self.indices)


class OutcomeUntransformer(TensorTransform):
    r"""Module acting as a bridge for `OutcomeTransform.untransform`."""

    def __init__(
        self,
        transform: OutcomeTransform,
        num_outputs: int | LongTensor,
    ):
        r"""Initializes an OutcomeUntransformer instance.

        Args:
            transform: The wrapped OutcomeTransform instance.
            num_outputs: The number of outcome features that the
                OutcomeTransform transforms.
        """
        super().__init__()
        self.transform = transform
        self.register_buffer(
            "num_outputs",
            num_outputs if torch.is_tensor(num_outputs) else torch.tensor(num_outputs),
        )

    def forward(self, values: Tensor) -> Tensor:
        # OutcomeTransforms expect an explicit output dimension in the final position.
        if self.num_outputs == 1:  # BoTorch has suppressed the output dimension
            output_values, _ = self.transform.untransform(values.unsqueeze(-1))
            return output_values.squeeze(-1)

        # BoTorch has moved the output dimension inside as the final batch dimension.
        output_values, _ = self.transform.untransform(values.transpose(-2, -1))
        return output_values.transpose(-2, -1)


def get_input_transform(model: GPyTorchModel) -> InputTransform | None:
    r"""Returns a model's input_transform or None."""
    return getattr(model, "input_transform", None)


def get_output_transform(model: GPyTorchModel) -> OutcomeUntransformer | None:
    r"""Returns a wrapped version of a model's outcome_transform or None."""
    transform = getattr(model, "outcome_transform", None)
    if transform is None:
        return None

    return OutcomeUntransformer(transform=transform, num_outputs=model.num_outputs)


@overload
def get_train_inputs(model: Model, transformed: bool = False) -> tuple[Tensor, ...]:
    pass  # pragma: no cover


@overload
def get_train_inputs(model: ModelList, transformed: bool = False) -> list[...]:
    pass  # pragma: no cover


def get_train_inputs(model: Model, transformed: bool = False):
    return GetTrainInputs(model, transformed=transformed)


@GetTrainInputs.register(Model)
def _get_train_inputs_Model(model: Model, transformed: bool = False) -> tuple[Tensor]:
    if not transformed:
        original_train_input = getattr(model, "_original_train_inputs", None)
        if torch.is_tensor(original_train_input):
            return (original_train_input,)

    (X,) = model.train_inputs
    transform = get_input_transform(model)
    if transform is None:
        return (X,)

    if model.training:
        return (transform.forward(X) if transformed else X,)
    return (X if transformed else transform.untransform(X),)


@GetTrainInputs.register(SingleTaskVariationalGP)
def _get_train_inputs_SingleTaskVariationalGP(
    model: SingleTaskVariationalGP, transformed: bool = False
) -> tuple[Tensor]:
    (X,) = model.model.train_inputs
    if model.training != transformed:
        return (X,)

    transform = get_input_transform(model)
    if transform is None:
        return (X,)

    return (transform.forward(X) if model.training else transform.untransform(X),)


@GetTrainInputs.register(ModelList)
def _get_train_inputs_ModelList(
    model: ModelList, transformed: bool = False
) -> list[...]:
    return [get_train_inputs(m, transformed=transformed) for m in model.models]


@overload
def get_train_targets(model: Model, transformed: bool = False) -> Tensor:
    pass  # pragma: no cover


@overload
def get_train_targets(model: ModelList, transformed: bool = False) -> list[...]:
    pass  # pragma: no cover


def get_train_targets(model: Model, transformed: bool = False):
    return GetTrainTargets(model, transformed=transformed)


@GetTrainTargets.register(Model)
def _get_train_targets_Model(model: Model, transformed: bool = False) -> Tensor:
    Y = model.train_targets

    # Note: Avoid using `get_output_transform` here since it creates a Module
    transform = getattr(model, "outcome_transform", None)
    if transformed or transform is None:
        return Y

    if model.num_outputs == 1:
        return transform.untransform(Y.unsqueeze(-1))[0].squeeze(-1)
    return transform.untransform(Y.transpose(-2, -1))[0].transpose(-2, -1)


@GetTrainTargets.register(SingleTaskVariationalGP)
def _get_train_targets_SingleTaskVariationalGP(
    model: Model, transformed: bool = False
) -> Tensor:
    Y = model.model.train_targets
    transform = getattr(model, "outcome_transform", None)
    if transformed or transform is None:
        return Y

    if model.num_outputs == 1:
        return transform.untransform(Y.unsqueeze(-1))[0].squeeze(-1)

    # SingleTaskVariationalGP.__init__ doesn't bring the multitoutpout dimension inside
    return transform.untransform(Y)[0]


@GetTrainTargets.register(ModelList)
def _get_train_targets_ModelList(
    model: ModelList, transformed: bool = False
) -> list[...]:
    return [get_train_targets(m, transformed=transformed) for m in model.models]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
References

.. [burt2020svgp]
    David R. Burt and Carl Edward Rasmussen and Mark van der Wilk,
    Convergence of Sparse Variational Inference in Gaussian Process Regression,
    Journal of Machine Learning Research, 2020,
    http://jmlr.org/papers/v21/19-1015.html.

.. [hensman2013svgp]
    James Hensman and Nicolo Fusi and Neil D. Lawrence, Gaussian Processes
    for Big Data, Proceedings of the 29th Conference on Uncertainty in
    Artificial Intelligence, 2013, https://arxiv.org/abs/1309.6835.

.. [moss2023ipa]
    Henry B. Moss and Sebastian W. Ober and Victor Picheny,
    Inducing Point Allocation for Sparse Gaussian Processes
    in High-Throughput Bayesian Optimization,Proceedings of
    the 25th International Conference on Artificial Intelligence
    and Statistics, 2023, https://arxiv.org/pdf/2301.10123.pdf.

"""

from __future__ import annotations

import copy
import warnings

import torch
from botorch.acquisition.objective import PosteriorTransform
from botorch.exceptions.warnings import UserInputWarning
from botorch.models.gpytorch import GPyTorchModel
from botorch.models.transforms.input import InputTransform
from botorch.models.transforms.outcome import OutcomeTransform
from botorch.models.utils import validate_input_scaling
from botorch.models.utils.gpytorch_modules import (
    get_covar_module_with_dim_scaled_prior,
    get_gaussian_likelihood_with_lognormal_prior,
)
from botorch.models.utils.inducing_point_allocators import (
    GreedyVarianceReduction,
    InducingPointAllocator,
)
from botorch.posteriors.gpytorch import GPyTorchPosterior
from gpytorch.distributions import MultivariateNormal
from gpytorch.kernels import Kernel
from gpytorch.likelihoods import (
    GaussianLikelihood,
    Likelihood,
    MultitaskGaussianLikelihood,
)
from gpytorch.means import ConstantMean, Mean
from gpytorch.models import ApproximateGP
from gpytorch.utils.memoize import clear_cache_hook
from gpytorch.variational import (
    _VariationalDistribution,
    _VariationalStrategy,
    CholeskyVariationalDistribution,
    IndependentMultitaskVariationalStrategy,
    VariationalStrategy,
)
from torch import Tensor
from torch.nn import Module
from typing_extensions import Self


TRANSFORM_WARNING = (
    "Using an {ttype} transform with `SingleTaskVariationalGP`. If this "
    "model is trained in minibatches, a {ttype} transform with learnable "
    "parameters would update its parameters for each minibatch, which is "
    "undesirable. If you do intend to train in minibatches, we recommend "
    "you not use a {ttype} transform and instead pre-transform your whole "
    "data set before fitting the model."
)


class ApproximateGPyTorchModel(GPyTorchModel):
    r"""
    Botorch wrapper class for various (variational) approximate GP models in
    GPyTorch.

    This can either include stochastic variational GPs (SVGPs) or
    variational implementations of weight space approximate GPs.
    """

    def __init__(
        self,
        model: ApproximateGP | None = None,
        likelihood: Likelihood | None = None,
        num_outputs: int = 1,
        *args,
        **kwargs,
    ) -> None:
        r"""
        Args:
            model: Instance of gpytorch.approximate GP models. If omitted,
                constructs a `_SingleTaskVariationalGP`.
            likelihood: Instance of a GPyTorch likelihood. If omitted, uses a
                either a `GaussianLikelihood` (if `num_outputs=1`) or a
                `MultitaskGaussianLikelihood`(if `num_outputs>1`).
            num_outputs: Number of outputs expected for the GP model.
            args: Optional positional arguments passed to the
                `_SingleTaskVariationalGP` constructor if no model is provided.
            kwargs: Optional keyword arguments passed to the
                `_SingleTaskVariationalGP` constructor if no model is provided.
        """
        super().__init__()

        self.model = (
            _SingleTaskVariationalGP(num_outputs=num_outputs, *args, **kwargs)
            if model is None
            else model
        )

        if likelihood is None:
            if num_outputs == 1:
                self.likelihood = GaussianLikelihood()
            else:
                self.likelihood = MultitaskGaussianLikelihood(num_tasks=num_outputs)
        else:
            self.likelihood = likelihood
        self._desired_num_outputs = num_outputs

    @property
    def num_outputs(self):
        return self._desired_num_outputs

    def eval(self) -> Self:
        r"""Puts the model in `eval` mode."""
        return Module.eval(self)

    def train(self, mode: bool = True) -> Self:
        r"""Put the model in `train` mode.

        Args:
            mode: A boolean denoting whether to put in `train` or `eval` mode.
                If `False`, model is put in `eval` mode.
        """
        return Module.train(self, mode=mode)

    def posterior(
        self,
        X,
        output_indices: list[int] | None = None,
        observation_noise: bool = False,
        posterior_transform: PosteriorTransform | None = None,
    ) -> GPyTorchPosterior:
        if output_indices is not None:
            raise NotImplementedError(  # pragma: no cover
                f"{self.__class__.__name__}.posterior does not support output indices."
            )
        self.eval()  # make sure model is in eval mode

        # input transforms are applied at `posterior` in `eval` mode, and at
        # `model.forward()` at the training time
        X = self.transform_inputs(X)

        # check for the multi-batch case for multi-outputs b/c this will throw
        # warnings
        X_ndim = X.ndim
        if self.num_outputs > 1 and X_ndim > 2:
            X = X.unsqueeze(-3).repeat(*[1] * (X_ndim - 2), self.num_outputs, 1, 1)
        dist = self.model(X)
        if observation_noise:
            dist = self.likelihood(dist)

        posterior = GPyTorchPosterior(distribution=dist)
        if hasattr(self, "outcome_transform"):
            posterior = self.outcome_transform.untransform_posterior(posterior, X=X)
        if posterior_transform is not None:
            posterior = posterior_transform(posterior)
        return posterior

    def forward(self, X) -> MultivariateNormal:
        if self.training:
            X = self.transform_inputs(X)
        return self.model(X)


class _SingleTaskVariationalGP(ApproximateGP):
    """
    Base class wrapper for a stochastic variational Gaussian Process (SVGP)
    model [hensman2013svgp]_.

    Uses by default pivoted Cholesky initialization for allocating inducing points,
    however, custom inducing point allocators can be provided.
    """

    def __init__(
        self,
        train_X: Tensor,
        train_Y: Tensor | None = None,
        num_outputs: int = 1,
        learn_inducing_points=True,
        covar_module: Kernel | None = None,
        mean_module: Mean | None = None,
        variational_distribution: _VariationalDistribution | None = None,
        variational_strategy: type[_VariationalStrategy] = VariationalStrategy,
        inducing_points: Tensor | int | None = None,
        inducing_point_allocator: InducingPointAllocator | None = None,
    ) -> None:
        r"""
        Args:
            train_X: Training inputs (due to the ability of the SVGP to sub-sample
                this does not have to be all of the training inputs).
            train_Y: Not used.
            num_outputs: Number of output responses per input.
            covar_module: Kernel function. If omitted, uses an `RBFKernel`.
            mean_module: Mean of GP model. If omitted, uses a `ConstantMean`.
            variational_distribution: Type of variational distribution to use
                (default: CholeskyVariationalDistribution), the properties of the
                variational distribution will encourage scalability or ease of
                optimization.
            variational_strategy: Type of variational strategy to use (default:
                VariationalStrategy). The default setting uses "whitening" of the
                variational distribution to make training easier.
            inducing_points: The number or specific locations of the inducing points.
            inducing_point_allocator: The `InducingPointAllocator` used to
                initialize the inducing point locations. If omitted,
                uses `GreedyVarianceReduction`.
        """
        # We use the model subclass wrapper to deal with input / outcome transforms.
        # The number of outputs will be correct here due to the check in
        # SingleTaskVariationalGP.
        input_batch_shape = train_X.shape[:-2]
        aug_batch_shape = copy.deepcopy(input_batch_shape)
        if num_outputs > 1:
            aug_batch_shape += torch.Size((num_outputs,))
        self._aug_batch_shape = aug_batch_shape

        if covar_module is None:
            covar_module = get_covar_module_with_dim_scaled_prior(
                ard_num_dims=train_X.shape[-1],
                batch_shape=self._aug_batch_shape,
            ).to(train_X)

        if inducing_point_allocator is None:
            inducing_point_allocator = GreedyVarianceReduction()

        # initialize inducing points if they are not given
        if not isinstance(inducing_points, Tensor):
            if inducing_points is None:
                # number of inducing points is 25% the number of data points
                # as a heuristic
                inducing_points = int(0.25 * train_X.shape[-2])

            inducing_points = inducing_point_allocator.allocate_inducing_points(
                inputs=train_X,
                covar_module=covar_module,
                num_inducing=inducing_points,
                input_batch_shape=input_batch_shape,
            )

        if variational_distribution is None:
            variational_distribution = CholeskyVariationalDistribution(
                num_inducing_points=inducing_points.shape[-2],
                batch_shape=self._aug_batch_shape,
            )

        variational_strategy_instance = variational_strategy(
            self,
            inducing_points=inducing_points,
            variational_distribution=variational_distribution,
            learn_inducing_locations=learn_inducing_points,
        )

        # wrap variational models in independent multi-task variational strategy
        if num_outputs > 1:
            variational_strategy_instance = IndependentMultitaskVariationalStrategy(
                base_variational_strategy=variational_strategy_instance,
                num_tasks=num_outputs,
                task_dim=-1,
            )
        super().__init__(variational_strategy=variational_strategy_instance)

        self.mean_module = (
            ConstantMean(batch_shape=self._aug_batch_shape).to(train_X)
            if mean_module is None
            else mean_module
        )

        self.covar_module = covar_module

    def forward(self, X) -> MultivariateNormal:
        mean_x = self.mean_module(X)
        covar_x = self.covar_module(X)
        latent_dist = MultivariateNormal(mean_x, covar_x)
        return latent_dist


class SingleTaskVariationalGP(ApproximateGPyTorchModel):
    r"""A single-task variational GP model following [hensman2013svgp]_.

    By default, the inducing points are initialized though the
    `GreedyVarianceReduction` of [burt2020svgp]_, which is known to be
    effective for building globally accurate models. However, custom
    inducing point allocators designed for specific down-stream tasks can also be
    provided (see [moss2023ipa]_ for details), e.g. `GreedyImprovementReduction`
    when the goal is to build a model suitable for standard BO.

    A single-task variational GP using relatively strong priors on the Kernel
    hyperparameters, which work best when covariates are normalized to the unit
    cube and outcomes are standardized (zero mean, unit variance).

    This model works in batch mode (each batch having its own hyperparameters).
    When the training observations include multiple outputs, this model will use
    batching to model outputs independently. However, batches of multi-output models
    are not supported at this time, if you need to use those, please use a
    ModelListGP.

    Use this model if you have a lot of data or if your responses are non-Gaussian.

    To train this model, you should use gpytorch.mlls.VariationalELBO and not
    the exact marginal log likelihood.

    Example:
        >>> import torch
        >>> from botorch.models import SingleTaskVariationalGP
        >>> from gpytorch.mlls import VariationalELBO
        >>>
        >>> train_X = torch.rand(20, 2)
        >>> model = SingleTaskVariationalGP(train_X)
        >>> mll = VariationalELBO(
        >>>     model.likelihood, model.model, num_data=train_X.shape[-2]
        >>> )
    """

    def __init__(
        self,
        train_X: Tensor,
        train_Y: Tensor | None = None,
        likelihood: Likelihood | None = None,
        num_outputs: int = 1,
        learn_inducing_points: bool = True,
        covar_module: Kernel | None = None,
        mean_module: Mean | None = None,
        variational_distribution: _VariationalDistribution | None = None,
        variational_strategy: type[_VariationalStrategy] = VariationalStrategy,
        inducing_points: Tensor | int | None = None,
        inducing_point_allocator: InducingPointAllocator | None = None,
        outcome_transform: OutcomeTransform | None = None,
        input_transform: InputTransform | None = None,
    ) -> None:
        r"""
        Args:
            train_X: Training inputs (due to the ability of the SVGP to sub-sample
                this does not have to be all of the training inputs).
            train_Y: Training targets (optional).
            likelihood: Instance of a GPyTorch likelihood. If omitted, uses a
                either a `GaussianLikelihood` (if `num_outputs=1`) or a
                `MultitaskGaussianLikelihood`(if `num_outputs>1`).
            num_outputs: Number of output responses per input (default: 1).
            learn_inducing_points: If True, the inducing point locations are learned
                jointly with the other model parameters.
            covar_module: Kernel function. If omitted, uses an `RBFKernel`.
            mean_module: Mean of GP model. If omitted, uses a `ConstantMean`.
            variational_distribution: Type of variational distribution to use
                (default: CholeskyVariationalDistribution), the properties of the
                variational distribution will encourage scalability or ease of
                optimization.
            variational_strategy: Type of variational strategy to use (default:
                VariationalStrategy). The default setting uses "whitening" of the
                variational distribution to make training easier.
            inducing_points: The number or specific locations of the inducing points.
            inducing_point_allocator: The `InducingPointAllocator` used to
                initialize the inducing point locations. If omitted,
                uses `GreedyVarianceReduction`.
            outcome_transform: An outcome transform that is applied to the training
                data during instantiation and to the posterior during inference.
                NOTE: If this model is trained in minibatches, an outcome transform
                with learnable parameters (such as `Standardize`) would update its
                parameters for each minibatch, which is undesirable. If you do intend
                to train in minibatches, we recommend you not use an outcome transform
                and instead pre-transform your whole data set before fitting the model.
            input_transform: An input transform that is applied in the model's
                forward pass.
                NOTE: If this model is trained in minibatches, an input transform
                with learnable parameters (such as `Normalize`) would update its
                parameters for each minibatch, which is undesirable. If you do intend
                to train in minibatches, we recommend you not use an input transform
                and instead pre-transform your whole data set before fitting the model.
        """
        with torch.no_grad():
            transformed_X = self.transform_inputs(
                X=train_X, input_transform=input_transform
            )
        if train_Y is not None:
            if outcome_transform is not None:
                warnings.warn(
                    TRANSFORM_WARNING.format(ttype="outcome"),
                    UserInputWarning,
                    stacklevel=3,
                )
                outcome_transform.train()
                train_Y, _ = outcome_transform(train_Y, X=transformed_X)
            self._validate_tensor_args(X=transformed_X, Y=train_Y)
            validate_input_scaling(
                train_X=transformed_X,
                train_Y=train_Y,
                check_nans_only=covar_module is not None,
            )
            if train_Y.shape[-1] != num_outputs:
                num_outputs = train_Y.shape[-1]

        self._num_outputs = num_outputs
        self._input_batch_shape = train_X.shape[:-2]
        aug_batch_shape = copy.deepcopy(self._input_batch_shape)
        if num_outputs > 1:
            aug_batch_shape += torch.Size([num_outputs])
        self._aug_batch_shape = aug_batch_shape

        if likelihood is None:
            if num_outputs == 1:
                likelihood = get_gaussian_likelihood_with_lognormal_prior(
                    batch_shape=self._aug_batch_shape
                )
            else:
                likelihood = MultitaskGaussianLikelihood(num_tasks=num_outputs)
        else:
            self._is_custom_likelihood = True

        if learn_inducing_points and (inducing_point_allocator is not None):
            warnings.warn(
                "After all the effort of specifying an inducing point allocator, "
                "you probably want to stop the inducing point locations "
                "being further optimized during the model fit. If so "
                "then set `learn_inducing_points` to False.",
                UserWarning,
                stacklevel=3,
            )

        if inducing_point_allocator is None:
            self._inducing_point_allocator = GreedyVarianceReduction()
        else:
            self._inducing_point_allocator = inducing_point_allocator

        model = _SingleTaskVariationalGP(
            train_X=transformed_X,
            num_outputs=num_outputs,
            learn_inducing_points=learn_inducing_points,
            covar_module=covar_module,
            mean_module=mean_module,
            variational_distribution=variational_distribution,
            variational_strategy=variational_strategy,
            inducing_points=inducing_points,
            inducing_point_allocator=self._inducing_point_allocator,
        )

        super().__init__(model=model, likelihood=likelihood, num_outputs=num_outputs)

        if outcome_transform is not None:
            self.outcome_transform = outcome_transform
        if input_transform is not None:
            warnings.warn(
                TRANSFORM_WARNING.format(ttype="input"),
                UserInputWarning,
                stacklevel=3,
            )
            self.input_transform = input_transform

        # for model fitting utilities
        # TODO: make this a flag?
        self.model.train_inputs = [transformed_X]
        if train_Y is not None:
            self.model.train_targets = train_Y.squeeze(-1)

        self.to(train_X)

    @property
    def batch_shape(self) -> torch.Size:
        r"""The batch shape of the model.

        This is a batch shape from an I/O perspective. For a model with `m`
        outputs, a `test_batch_shape x q x d`-shaped input `X` to the `posterior`
        method returns a Posterior object over an output of shape
        `broadcast(test_batch_shape, model.batch_shape) x q x m`.
        """
        return self._input_batch_shape

    def init_inducing_points(
        self,
        inputs: Tensor,
    ) -> Tensor:
        r"""
        Reinitialize the inducing point locations in-place with the current kernel
        applied to `inputs` through the model's inducing point allocation strategy.
        The variational distribution and variational strategy caches are reset.

        Args:
            inputs: (\*batch_shape, n, d)-dim input data tensor.

        Returns:
            (\*batch_shape, m, d)-dim tensor of selected inducing point locations.
        """
        var_strat = self.model.variational_strategy
        clear_cache_hook(var_strat)
        if hasattr(var_strat, "base_variational_strategy"):
            var_strat = var_strat.base_variational_strategy
            clear_cache_hook(var_strat)

        with torch.no_grad():
            num_inducing = var_strat.inducing_points.size(-2)
            inducing_points = self._inducing_point_allocator.allocate_inducing_points(
                inputs=inputs,
                covar_module=self.model.covar_module,
                num_inducing=num_inducing,
                input_batch_shape=self._input_batch_shape,
            )
            var_strat.inducing_points.copy_(inducing_points)
            var_strat.variational_params_initialized.fill_(0)

        return inducing_points



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
This file contains an implemenation of a Variational Bayesian Last Layer (VBLL) model
that can be used within BoTorch for Bayesian optimization.

References:

[1] P. Brunzema, M. Jordahn, J. Willes, S. Trimpe, J. Snoek, J. Harrison.
    Bayesian Optimization via Contrinual Variational Last Layer Training.
    International Conference on Learning Representations, 2025.

Contributor: brunzema
"""

from __future__ import annotations

import torch
import torch.nn as nn

from botorch.logging import logger
from botorch.posteriors import Posterior
from botorch_community.models.blls import AbstractBLLModel
from botorch_community.models.vbll_helper import DenseNormal, Normal, Regression
from botorch_community.posteriors.bll_posterior import BLLPosterior
from gpytorch.distributions import MultivariateNormal
from torch import Tensor
from torch.optim import Optimizer
from torch.utils.data import DataLoader


class SampleModel(nn.Module):
    def __init__(
        self,
        backbone: nn.Module,
        sampled_params: Tensor,
        num_inputs: int,
        num_outputs: int,
    ):
        """Network for posterior samples of BLL models.

        Args:
            backbone: Backbone of the original model BLL model
            sampled_params: Sampled parameters from the posterior distribution
                of the last layer.
            num_inputs: Number of inputs to the backbone.
            num_outputs: Number of outputs
        """
        super().__init__()
        self.backbone = backbone
        self.sampled_params = sampled_params
        self.num_inputs = num_inputs
        self.num_outputs = num_outputs

    def forward(self, x: Tensor) -> Tensor:
        """Forward pass through the sample network.

        Args:
            x: Input as `batch_size x d`-dim Tensor.

        Returns:
            Output as `(sample_shape), batch_size, output_dim`-dim Tensor.
        """
        x = self.backbone(x)

        if self.sampled_params.dim() == 2:
            return (self.sampled_params @ x[..., None]).squeeze(-1)

        x_expanded = x.unsqueeze(0).expand(self.sampled_params.shape[0], -1, -1)
        return x_expanded @ self.sampled_params.transpose(-1, -2)


class VBLLNetwork(nn.Module):
    def __init__(
        self,
        in_features: int = None,
        hidden_features: int = 64,
        out_features: int = 1,
        num_layers: int = 3,
        parameterization: str = "dense",
        cov_rank: int | None = None,
        mean_initialization: str | None = None,
        prior_scale: float = 1.0,
        wishart_scale: float = 0.01,
        clamp_noise_init: bool = True,
        kl_scale: float = 1.0,
        backbone: nn.Module | None = None,
        activation: nn.Module = nn.ELU(),
        device=None,
    ):
        """
        A model with a Variational Bayesian Linear Last (VBLL) layer.

        Args:
            in_features: Number of input features. Defaults to 2.
            hidden_features: Number of hidden units per layer. Defaults to 50.
            out_features: Number of output features. Defaults to 1.
            num_layers: Number of hidden layers in the MLP. Defaults to 3.
            parameterization: Parameterization of the posterior covariance of the last
                layer. Supports {'dense', 'diagonal', 'lowrank', 'dense_precision'}.
            cov_rank: For 'lowrank' parameterization, the rank of the covariance matrix.
            mean_initialization: Initialization method for the mean of the weights in
                the last layer. Supports {'kaiming', None}. If None, weights are
                initialized from a standard normal distribution. Defaults to None.
            prior_scale: Scaling factor for the prior distribution in the Bayesian last
                layer. Defaults to 1.0.
            wishart_scale: Scaling factor for the Wishart prior in the Bayesian last
                layer. Defaults to 0.01.
            kl_scale: Weighting factor for the Kullback-Leibler (KL) divergence term in
                the loss. Defaults to 1.0.
            backbone: A predefined feature extractor to be used before the VBLL layer.
                If None, a default MLP structure is used. Defaults to None.
            activation: Activation function applied between hidden layers.
                Defaults to `nn.ELU()`.
            device: The device on which the model is stored. If None, the device is
                inferred based on availability. Defaults to None.

        Notes:
            - If a `backbone` module is provided, it is applied before the variational
            last layer. If not, we use a default MLP structure.
        """
        super().__init__()
        self.num_outputs = out_features

        if device is None:
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self.device = device

        self.activation = activation
        self.kl_scale = kl_scale

        if backbone is None:
            if in_features is None:
                raise ValueError(
                    "Please specify the input dimension in the constructor."
                )

            hidden_layers = [
                nn.Sequential(
                    nn.Linear(hidden_features, hidden_features),
                    self.activation,
                )
                for _ in range(num_layers)
            ]
            self.backbone = nn.Sequential(
                nn.Linear(in_features, hidden_features),
                self.activation,
                *hidden_layers,
            ).to(dtype=torch.float64, device=self.device)
            self.num_inputs = in_features

        else:
            self.backbone = backbone

            # Try to infer input size if backbone is a Sequential and starts with Linear
            if isinstance(backbone, nn.Sequential) and isinstance(
                backbone[0], nn.Linear
            ):
                self.num_inputs = backbone[0].in_features
            elif in_features is not None:
                self.num_inputs = in_features
            else:
                raise ValueError(
                    "Cannot infer input dimension from provided backbone. "
                    "Please specify `in_features` explicitly."
                )

        # could be changed to other vbll regression layers
        self.head = Regression(
            hidden_features,
            out_features,
            regularization_weight=1.0,  # will be adjusted dynamically
            prior_scale=prior_scale,
            wishart_scale=wishart_scale,
            parameterization=parameterization,
            cov_rank=cov_rank,
            clamp_noise_init=clamp_noise_init,
            mean_initialization=mean_initialization,
        ).to(dtype=torch.float64, device=self.device)

    def forward(self, x: Tensor) -> Tensor:
        """Forward pass through the VBLL network.

        Args:
            x: Input as `batch_shape x in_features`-dim Tensor.

        Returns:
            Output as `batch_shape x out_features`-dim Tensor.
        """
        x = self.backbone(x)
        return self.head(x)

    def sample_posterior_function(
        self, sample_shape: torch.Size | None = None
    ) -> nn.Module:
        """
        Samples a posterior function by drawing parameters from the model's learned
        distribution.

        Args:
            sample_shape: The desired shape for the sampled parameters. If None, a
                single sample is drawn. Defaults to None.

        Returns:
            A nn.Module that takes an input tensor `x` and returns the corresponding
            model output tensor. The function applies the backbone transformation
            and computes the final output using the sampled parameters.

        Notes:
            - If `sample_shape` is provided, multiple samples are drawn, and the
            function will return a batched output where the first dimension corresponds
            to different samples.
        """
        sample_shape = sample_shape or torch.Size()
        sampled_params = self.head.W().rsample(sample_shape).to(self.device)
        return SampleModel(
            backbone=self.backbone,
            sampled_params=sampled_params,
            num_inputs=self.num_inputs,
            num_outputs=self.num_outputs,
        )


def _get_optimizer(
    optimizer_class: Optimizer,
    model_parameters: list,
    lr: float = 1e-3,
    **kwargs,
) -> Optimizer:
    """
    Creates and returns an optimizer.

    Args:
        optimizer_class: The optimizer class (e.g., torch.optim.AdamW).
        model_parameters: Parameters to be optimized.
        lr: Learning rate.
        **kwargs: Additional arguments to be passed to the optimizer.

    Returns:
        The initialized optimizer.
    """
    return optimizer_class(model_parameters, lr=lr, **kwargs)


class VBLLModel(AbstractBLLModel):
    def __init__(self, *args, **kwargs):
        """BoTorch compatible VBLL model."""
        super().__init__()
        self.model = VBLLNetwork(*args, **kwargs)

    @property
    def backbone(self):
        return self.model.backbone

    @property
    def head(self):
        return self.model.head

    def sample(self, sample_shape: torch.Size | None = None) -> nn.Module:
        """Create posterior sample networks of the VBLL model. Note that posterior
        samples, we first sample from the posterior distribution of the last layer and
        then create a generalized linear model to get the posterior samples.

        Args:
            sample_shape: Number of samples to draw from the posterior distribution.
                If None, a single sample is drawn. Defaults to None.

        Returns:
            A nn.Module that takes an input as `batch_size x num_inputs`-dim Tensor and
            returns a `(sample_shape), batch_size, output_dim`-dim Tensor.
        """
        return self.model.sample_posterior_function(sample_shape)

    def __call__(self, X: Tensor) -> Normal | DenseNormal:
        """Forward pass through the VBLL model.

        Args:
            X: Input as `batch_size x num_inputs`-dim Tensor.

        Returns:
            Normal distribution with `batch_size x num_outputs` as the mean and
            `batch_size x num_outputs` as variance.
        """
        return self.model(X).predictive

    def fit(
        self,
        train_X: Tensor,
        train_y: Tensor,
        val_X: Tensor | None = None,
        val_y: Tensor | None = None,
        optimization_settings: dict | None = None,
        initialization_params: dict | None = None,
    ):
        """
        Fits the model to the given training data.

        Args:
            train_X: The input training data, expected to be a Tensor of shape
                `num_samples x d`.

            train_y: The target values for training, expected to be a Tensor of
                shape `num_samples x num_outputs`.

            val_X: The optional input validation data, expected to be a Tensor of shape
                `num_val_samples x d`.

            val_y: The optional target values for validation, expected to be a Tensor of
                shape `num_val_samples x num_outputs`.

            optimization_settings: A dict containing optimization-related settings.
                If a key is missing, default values will be used. Available settings:
                    - "num_epochs" (int, default=100): The maximum number of epochs.
                    - "patience" (int, default=10): Epochs before early stopping.
                    - "freeze_backbone" (bool, default=False): If True, the backbone of
                    the model is frozen.
                    - "batch_size" (int, default=32): Batch size for the training.
                    - "optimizer" (torch.optim.Optimizer, default=torch.optim.AdamW):
                    Optimizer for training.
                    - "wd" (float, default=1e-4): Weight decay (L2 regularization)
                    coefficient.
                    - "clip_val" (float, default=1.0): Gradient clipping threshold.

            initialization_params: A dictionary containing the initial parameters of the
                model for feature reuse. If None, the optimization will start from from
                the random initialization in the __init__ method.

        Returns:
            The function trains the model in place and does not return a value.
        """

        if (val_X is None) != (val_y is None):
            missing = "val_X" if val_X is None else "val_y"
            provided = "val_y" if val_X is None else "val_X"

            raise ValueError(
                f"Validation error: {missing} is None while {provided} is provided. "
                "Either both validation inputs (val_X and val_y) must be provided, or"
                "neither."
            )

        # Default settings
        default_opt_settings = {
            "num_epochs": 10_000,
            "freeze_backbone": False,
            "patience": 100,
            "batch_size": 32,
            "optimizer_class": torch.optim.AdamW,
            "lr": 1e-3,
            "wd": 1e-4,
            "clip_val": 1.0,
            "optimizer_kwargs": {},  # Optimizer-specific args (e.g., betas for Adam)
        }

        # Merge defaults with provided settings
        optimization_settings = {
            **default_opt_settings,
            **(optimization_settings or {}),
        }

        # Make dataloader based on train_X, train_y
        device = self.model.device
        dataset = [[train_X[i], train_y[i]] for i, _ in enumerate(train_X)]

        dataloader = DataLoader(
            dataset, shuffle=True, batch_size=optimization_settings["batch_size"]
        )

        if initialization_params is not None:
            self.model.load_state_dict(initialization_params)

        self.model.to(device)
        self.set_reg_weight(self.model.kl_scale / len(train_y))
        param_list = [
            {
                "params": self.model.head.parameters(),
                "weight_decay": 0.0,
            },
        ]

        # freeze backbone
        if not optimization_settings["freeze_backbone"]:
            param_list.append(
                {
                    "params": self.model.backbone.parameters(),
                    "weight_decay": optimization_settings["wd"],
                }
            )

        # Extract settings
        optimizer_class = optimization_settings["optimizer_class"]
        optimizer_kwargs = optimization_settings.get("optimizer_kwargs", {})

        # Initialize optimizer using helper function
        optimizer = _get_optimizer(
            optimizer_class=optimizer_class,
            model_parameters=param_list,
            lr=optimization_settings["lr"],
            **optimizer_kwargs,
        )

        best_loss = float("inf")
        epochs_no_improve = 0
        early_stop = False
        best_model_state = None  # To store the best model parameters

        self.model.train()

        for epoch in range(1, optimization_settings["num_epochs"] + 1):
            # early stopping
            if early_stop:
                break

            running_loss = []

            for train_step, (x, y) in enumerate(dataloader):
                x, y = x.to(device), y.to(device)
                optimizer.zero_grad()
                out = self.model(x)
                loss = out.train_loss_fn(y)  # vbll layer will calculate the loss
                loss.backward()

                if optimization_settings["clip_val"] is not None:
                    torch.nn.utils.clip_grad_norm_(
                        self.model.parameters(), optimization_settings["clip_val"]
                    )

                optimizer.step()
                running_loss.append(loss.item())

            # Calculate average loss over the epoch
            avg_train_loss = sum(running_loss[-len(dataloader) :]) / len(dataloader)

            # If validation data is provided, compute validation loss
            if val_X is not None and val_y is not None:
                self.model.eval()  # Set model to evaluation mode
                with torch.no_grad():
                    out = self.model(x)
                    val_loss = out.val_loss_fn(val_y)

                self.model.train()  # Set model back to training mode

                # Use validation loss for early stopping
                current_loss = val_loss.item()
            else:
                # If no validation data, use training loss
                current_loss = avg_train_loss

            # Early stopping logic
            if current_loss < best_loss:
                best_loss = current_loss
                best_model_state = self.model.state_dict()
                epochs_no_improve = 0
            else:
                epochs_no_improve += 1

            if epochs_no_improve >= optimization_settings["patience"]:
                early_stop = True

        # load best model
        if best_model_state is not None:
            self.model.load_state_dict(best_model_state)
            logger.info(f"Early stopping at epoch {epoch} with loss {best_loss}.")

    def set_reg_weight(self, new_weight: float):
        self.model.head.regularization_weight = new_weight

    def posterior(
        self,
        X: Tensor,
        output_indices=None,
        observation_noise=None,
        posterior_transform=None,
    ) -> Posterior:
        if X.dim() > 3:
            raise ValueError(f"Input must have at most 3 dimensions, got {X.dim()}.")

        # Determine if the input is batched
        batched = X.dim() == 3

        if not batched:
            N, D = X.shape
            B = 1
        else:
            B, N, D = X.shape
            X = X.reshape(B * N, D)

        posterior = self.model(X).predictive

        # Extract mean and variance
        mean = posterior.mean.squeeze(dim=-1)
        variance = posterior.variance.squeeze(dim=-1)
        cov = torch.diag_embed(variance)

        K = self.num_outputs
        mean = mean.reshape(B, N * K)

        # Cov must be `(B, N*K, N*K)`
        cov = cov.reshape(B, N, K, B, N, K)
        cov = torch.einsum("bnkbrl->bnkrl", cov)  # (B, N, K, N, K)
        cov = cov.reshape(B, N * K, N * K)

        # Remove fake batch dimension if not batched
        if not batched:
            mean = mean.squeeze(0)
            cov = cov.squeeze(0)

        # pass as MultivariateNormal to BLLPosterior
        distribution = MultivariateNormal(mean, cov)
        return BLLPosterior(
            model=self, distribution=distribution, X=X, output_dim=self.num_outputs
        )

    def __str__(self) -> str:
        return self.model.__str__()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any

import torch
from botorch.models.utils.gpytorch_modules import get_covar_module_with_dim_scaled_prior
from gpytorch.constraints import Positive
from gpytorch.kernels.kernel import Kernel
from gpytorch.priors.torch_priors import GammaPrior
from linear_operator.operators import DiagLinearOperator
from linear_operator.operators.dense_linear_operator import DenseLinearOperator
from torch import Tensor
from torch.nn import ModuleList


def get_order(indices: list[int]) -> list[int]:
    r"""Get the order indices as integers ranging from 0 to the number of indices.

    Args:
        indices: A list of parameter indices.

    Returns:
        A list of integers ranging from 0 to the number of indices.
    """
    return [i % len(indices) for i in indices]


def is_contiguous(indices: list[int]) -> bool:
    r"""Check if the list of integers is contiguous.

    Args:
        indices: A list of parameter indices.
    Returns:
        A boolean indicating whether the indices are contiguous.
    """
    min_idx = min(indices)
    return set(indices) == set(range(min_idx, min_idx + len(indices)))


def get_permutation(decomposition: dict[str, list[int]]) -> list[int] | None:
    """Construct permutation to reorder the parameters such that:

    1) the parameters for each context are contiguous.
    2) The parameters for each context are in the same order

    Args:
        decomposition: A dictionary mapping context names to a list of
            parameters.
    Returns:
        A permutation to reorder the parameters for (1) and (2).
        Returning `None` means that ordering specified in `decomposition`
        satisfies (1) and (2).
    """
    permutation = None
    if not all(
        is_contiguous(indices=active_parameters)
        for active_parameters in decomposition.values()
    ):
        permutation = _create_new_permutation(decomposition=decomposition)
    else:
        same_order = True
        expected_order = get_order(indices=next(iter(decomposition.values())))
        for active_parameters in decomposition.values():
            order = get_order(indices=active_parameters)
            if order != expected_order:
                same_order = False
                break
        if not same_order:
            permutation = _create_new_permutation(decomposition=decomposition)
    return permutation


def _create_new_permutation(decomposition: dict[str, list[int]]) -> list[int]:
    # make contiguous and ordered
    permutation = []
    for active_parameters in decomposition.values():
        sorted_indices = sorted(active_parameters)
        permutation.extend(sorted_indices)
    return permutation


class LCEAKernel(Kernel):
    r"""The Latent Context Embedding Additive (LCE-A) Kernel.

    This kernel is similar to the SACKernel, and is used when context breakdowns are
    unbserverable. It assumes the same additive structure and a spatial kernel shared
    across contexts. Rather than assuming independence, LCEAKernel models the
    correlation in the latent functions for each context through learning context
    embeddings.
    """

    def __init__(
        self,
        decomposition: dict[str, list[int]],
        batch_shape: torch.Size,
        train_embedding: bool = True,
        cat_feature_dict: dict | None = None,
        embs_feature_dict: dict | None = None,
        embs_dim_list: list[int] | None = None,
        context_weight_dict: dict | None = None,
        device: torch.device | None = None,
    ) -> None:
        r"""
        Args:
            decomposition: Keys index context names. Values are the indexes of
                parameters belong to the context.
            batch_shape: Batch shape as usual for gpytorch kernels. Model does not
                support batch training. When batch_shape is non-empty, it is used for
                loading hyper-parameter values generated from MCMC sampling.
            train_embedding: A boolean indictor of whether to learn context embeddings.
            cat_feature_dict: Keys are context names and values are list of categorical
                features i.e. {"context_name" : [cat_0, ..., cat_k]}. k equals the
                number of categorical variables. If None, uses context names in the
                decomposition as the only categorical feature, i.e., k = 1.
            embs_feature_dict: Pre-trained continuous embedding features of each
                context.
            embs_dim_list: Embedding dimension for each categorical variable. The length
                equals to num of categorical features k. If None, the embedding
                dimension is set to 1 for each categorical variable.
            context_weight_dict: Known population weights of each context.
        """
        super().__init__(batch_shape=batch_shape)
        self.batch_shape = batch_shape
        self.train_embedding = train_embedding
        self._device = device

        self.num_param = len(next(iter(decomposition.values())))
        self.context_list = list(decomposition.keys())
        self.num_contexts = len(self.context_list)

        # get parameter space decomposition
        for active_parameters in decomposition.values():
            # check number of parameters are same in each decomp
            if len(active_parameters) != self.num_param:
                raise ValueError(
                    "The number of parameters needs to be same across all contexts."
                )
        # reorder the parameter list based on decomposition such that
        # parameters for each context are contiguous and in the same order for each
        # context
        self.permutation = get_permutation(decomposition=decomposition)
        # get context features and set emb dim
        self.context_cat_feature = None
        self.context_emb_feature = None
        self.n_embs = 0
        self.emb_weight_matrix_list = None
        self.emb_dims = None
        self._set_context_features(
            cat_feature_dict=cat_feature_dict,
            embs_feature_dict=embs_feature_dict,
            embs_dim_list=embs_dim_list,
        )
        # contruct embedding layer
        if train_embedding:
            self._set_emb_layers()
        # task covariance matrix
        self.task_covar_module = get_covar_module_with_dim_scaled_prior(
            ard_num_dims=self.n_embs,
            batch_shape=batch_shape,
        )
        # base kernel
        self.base_kernel = get_covar_module_with_dim_scaled_prior(
            ard_num_dims=self.num_param,
            batch_shape=batch_shape,
        )
        # outputscales for each context (note this is like sqrt of outputscale)
        self.context_weight = None
        if context_weight_dict is None:
            outputscale_list = torch.zeros(
                *batch_shape, self.num_contexts, device=self.device
            )
        else:
            outputscale_list = torch.zeros(*batch_shape, 1, device=self.device)
            self.context_weight = torch.tensor(
                [context_weight_dict[c] for c in self.context_list], device=self.device
            )
        self.register_parameter(
            name="raw_outputscale_list", parameter=torch.nn.Parameter(outputscale_list)
        )
        self.register_prior(
            "outputscale_list_prior",
            GammaPrior(2.0, 15.0),
            lambda m: m.outputscale_list,
            lambda m, v: m._set_outputscale_list(v),
        )
        self.register_constraint("raw_outputscale_list", Positive())

    @property
    def device(self) -> torch.device | None:
        return self._device

    @property
    def outputscale_list(self) -> Tensor:
        return self.raw_outputscale_list_constraint.transform(self.raw_outputscale_list)

    @outputscale_list.setter
    def outputscale_list(self, value: Tensor) -> None:
        self._set_outputscale_list(value)

    def _set_outputscale_list(self, value: Tensor) -> None:
        if not torch.is_tensor(value):
            value = torch.as_tensor(value).to(self.raw_outputscale_list)
        self.initialize(
            raw_outputscale_list=self.raw_outputscale_list_constraint.inverse_transform(
                value
            )
        )

    def _set_context_features(
        self,
        cat_feature_dict: dict | None = None,
        embs_feature_dict: dict | None = None,
        embs_dim_list: list[int] | None = None,
    ) -> None:
        """Set context categorical features and continuous embedding features.
        If cat_feature_dict is None, context indices will be used; If embs_dim_list
        is None, we use 1-d embedding for each categorical features.
        """
        # get context categorical features
        if cat_feature_dict is None:
            self.context_cat_feature = torch.arange(
                self.num_contexts, device=self.device
            ).unsqueeze(-1)
        else:
            self.context_cat_feature = torch.tensor(
                [cat_feature_dict[c] for c in self.context_list]
            )
        #  construct emb_dims based on categorical features
        if embs_dim_list is None:
            #  set embedding_dim = 1 for each categorical variable
            embs_dim_list = [1 for _i in range(self.context_cat_feature.size(1))]
        self.emb_dims = [
            (len(self.context_cat_feature[:, i].unique()), embs_dim_list[i])
            for i in range(self.context_cat_feature.size(1))
        ]
        if self.train_embedding:
            self.n_embs = sum(embs_dim_list)  # total num of emb features
        # get context embedding features
        if embs_feature_dict is not None:
            self.context_emb_feature = torch.tensor(
                [embs_feature_dict[c] for c in self.context_list], device=self.device
            )
            self.n_embs += self.context_emb_feature.size(1)

    def _set_emb_layers(self) -> None:
        """Construct embedding layers.
        If model is non-batch, we use nn.Embedding to learn emb weights. If model is
        batched (sef.batch_shape is non-empty), we load emb weights posterior samples
        and construct a parameter list that each parameter is the emb weight of each
        layer. The shape of weight matrices are ns x num_contexts x emb_dim.
        """
        self.emb_layers = ModuleList(
            [
                torch.nn.Embedding(num_embeddings=x, embedding_dim=y, max_norm=1.0)
                for x, y in self.emb_dims
            ]
        )
        # use posterior of emb weights
        if len(self.batch_shape) > 0:
            self.emb_weight_matrix_list = torch.nn.ParameterList(
                [
                    torch.nn.Parameter(
                        torch.zeros(
                            self.batch_shape + emb_layer.weight.shape,
                            device=self.device,
                        )
                    )
                    for emb_layer in self.emb_layers
                ]
            )

    def _eval_context_covar(self) -> Tensor:
        """Compute context covariance matrix.

        Returns:
            A (ns) x num_contexts x num_contexts tensor.
        """
        if len(self.batch_shape) > 0:
            # broadcast - (ns x num_contexts x k)
            all_embs = self._task_embeddings_batch()
        else:
            all_embs = self._task_embeddings()  # no broadcast - (num_contexts x k)

        context_covar = self.task_covar_module(all_embs).to_dense()
        if self.context_weight is None:
            context_outputscales = self.outputscale_list
        else:
            context_outputscales = self.outputscale_list * self.context_weight
        context_covar = (
            (context_outputscales.unsqueeze(-2))
            # (ns) x 1 x num_contexts
            .mul(context_covar)
            .mul(context_outputscales.unsqueeze(-1))  # (ns) x num_contexts x 1
        )
        return context_covar

    def _task_embeddings(self) -> Tensor:
        """Generate embedding features of contexts when model is non-batch.

        Returns:
            a (num_contexts x n_embs) tensor. n_embs is the sum of embedding
            dimensions i.e. sum(embs_dim_list)
        """
        if self.train_embedding is False:
            return self.context_emb_feature  # use pre-trained embedding only
        context_features = torch.stack(
            [self.context_cat_feature[i, :] for i in range(self.num_contexts)], dim=0
        )
        embeddings = [
            emb_layer(context_features[:, i].to(device=self.device, dtype=torch.long))
            for i, emb_layer in enumerate(self.emb_layers)
        ]
        embeddings = torch.cat(embeddings, dim=1)
        # add given embeddings if any
        if self.context_emb_feature is not None:
            embeddings = torch.cat([embeddings, self.context_emb_feature], dim=1)
        return embeddings

    def _task_embeddings_batch(self) -> Tensor:
        """Generate embedding features of contexts when model has multiple batches.

        Returns:
            a (ns) x num_contexts x n_embs tensor. ns is the batch size i.e num of
            posterior samples; n_embs is the sum of embedding dimensions i.e.
            sum(embs_dim_list).
        """
        context_features = torch.cat(
            [
                self.context_cat_feature[i, :].unsqueeze(0)
                for i in range(self.num_contexts)
            ]
        )
        embeddings = []
        for b in range(self.batch_shape.numel()):  # pyre-ignore
            for i in range(len(self.emb_weight_matrix_list)):
                # loop over emb layer and concat embs from each layer
                embeddings.append(
                    torch.cat(
                        [
                            torch.nn.functional.embedding(
                                context_features[:, 0].to(
                                    dtype=torch.long, device=self.device
                                ),
                                self.emb_weight_matrix_list[i][b, :],
                            ).unsqueeze(0)
                        ],
                        dim=1,
                    )
                )
        embeddings = torch.cat(embeddings, dim=0)
        # add given embeddings if any
        if self.context_emb_feature is not None:
            embeddings = torch.cat(
                [
                    embeddings,
                    self.context_emb_feature.expand(
                        *self.batch_shape + self.context_emb_feature.shape
                    ),
                ],
                dim=-1,
            )
        return embeddings

    def train(self, mode: bool = True) -> None:
        super().train(mode=mode)
        if not mode:
            self.register_buffer("_context_covar", self._eval_context_covar())

    def forward(
        self,
        x1: Tensor,
        x2: Tensor,
        diag: bool = False,
        last_dim_is_batch: bool = False,
        **params: Any,
    ) -> Tensor:
        """Iterate across each partition of parameter space and sum the
        covariance matrices together
        """
        # context covar matrix
        context_covar = (
            self._eval_context_covar() if self.training else self._context_covar
        )
        base_covar_perm = self._eval_base_covar_perm(x1, x2)
        # expand context_covar to match base_covar_perm
        if base_covar_perm.dim() > context_covar.dim():
            context_covar = context_covar.expand(base_covar_perm.shape)
        # then weight by the context kernel
        # compute the base kernel on the d parameters
        einsum_str = "...nnki, ...nnki -> ...n" if diag else "...ki, ...ki -> ..."
        covar_dense = torch.einsum(einsum_str, context_covar, base_covar_perm)
        if diag:
            return DiagLinearOperator(covar_dense)
        return DenseLinearOperator(covar_dense)

    def _eval_base_covar_perm(self, x1: Tensor, x2: Tensor) -> Tensor:
        """Computes the base covariance matrix on x1, x2, applying permutations and
        reshaping the kernel matrix as required by `forward`.

        NOTE: Using the notation n = num_observations, k = num_contexts, d = input_dim,
        the input tensors have to have the following shapes.

        Args:
            x1: `batch_shape x n x (k*d)`-dim Tensor of kernel inputs.
            x2: `batch_shape x n x (k*d)`-dim Tensor of kernel inputs.

        Returns:
            `batch_shape x n x n x k x k`-dim Tensor of base covariance values.
        """
        if self.permutation is not None:
            x1 = x1[..., self.permutation]
            x2 = x2[..., self.permutation]
        # turn last two dimensions of n x (k*d) into (n*k) x d.
        x1_exp = x1.reshape(*x1.shape[:-2], -1, self.num_param)
        x2_exp = x2.reshape(*x2.shape[:-2], -1, self.num_param)
        # batch shape x n*k x n*k
        base_covar = self.base_kernel(x1_exp, x2_exp)
        # batch shape x n x n x k x k
        view_shape = x1.shape[:-2] + torch.Size(
            [
                x1.shape[-2],
                self.num_contexts,
                x2.shape[-2],
                self.num_contexts,
            ]
        )
        base_covar_perm = (
            base_covar.to_dense()
            .view(view_shape)
            .permute(*list(range(x1.ndim - 2)), -4, -2, -3, -1)
        )
        return base_covar_perm



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

from collections.abc import Callable

from warnings import warn

import torch
from botorch.posteriors.gpytorch import GPyTorchPosterior
from gpytorch.distributions import MultivariateNormal
from torch import Tensor


MCMC_DIM = -3  # Location of the MCMC batch dimension
TOL = 1e-6  # Bisection tolerance


def batched_bisect(
    f: Callable, target: float, bounds: Tensor, tol: float = TOL, max_steps: int = 32
):
    r"""Batched bisection with a fixed number of steps.

    Args:
        f: Target function that takes a `(b1 x ... x bk)`-dim tensor and returns a
            `(b1 x ... x bk)`-dim tensor.
        target: Scalar target value of type float.
        bounds: Lower and upper bounds, of size `2 x b1 x ... x bk`.
        tol: We termniate if all elements satisfy are within `tol` of the `target`.
        max_steps: Maximum number of bisection steps.

    Returns:
        Tensor X of size `b1 x ... x bk` such that `f(X) = target`.
    """
    # Make sure target is actually contained in the interval
    f1, f2 = f(bounds[0]), f(bounds[1])
    if not ((f1 <= target) & (target <= f2)).all():
        raise ValueError(
            "The target is not contained in the interval specified by the bounds"
        )
    bounds = bounds.clone()  # Will be modified in-place
    center = bounds.mean(dim=0)
    f_center = f(center)
    for _ in range(max_steps):
        go_left = f_center > target
        bounds[1, go_left] = center[go_left]
        bounds[0, ~go_left] = center[~go_left]
        center = bounds.mean(dim=0)
        f_center = f(center)
        # Check convergence
        if (f_center - target).abs().max() <= tol:
            return center
    return center


def _quantile(posterior: GaussianMixturePosterior, value: Tensor) -> Tensor:
    r"""Compute the posterior quantiles for the mixture of models."""
    if value.numel() > 1:
        return torch.stack(
            [_quantile(posterior=posterior, value=v) for v in value], dim=0
        )
    if value <= 0 or value >= 1:
        raise ValueError("value is expected to be in the range (0, 1).")
    dist = torch.distributions.Normal(
        loc=posterior.mean, scale=posterior.variance.sqrt()
    )
    if posterior.mean.shape[MCMC_DIM] == 1:  # Analytical solution
        return dist.icdf(value).squeeze(MCMC_DIM)
    icdf_val = dist.icdf(value)
    low = icdf_val.min(dim=MCMC_DIM).values - TOL
    high = icdf_val.max(dim=MCMC_DIM).values + TOL
    bounds = torch.cat((low.unsqueeze(0), high.unsqueeze(0)), dim=0)
    return batched_bisect(
        f=lambda x: dist.cdf(x.unsqueeze(MCMC_DIM)).mean(dim=MCMC_DIM),
        target=value.item(),
        bounds=bounds,
    )


class GaussianMixturePosterior(GPyTorchPosterior):
    r"""A Gaussian mixture posterior.

    The MCMC batch dimension that corresponds to the models in the mixture is located
    at `MCMC_DIM` (defined at the top of this file). Note that while each MCMC sample
    corresponds to a Gaussian posterior, the posterior is rather a mixture of Gaussian
    distributions.
    """

    def __init__(self, distribution: MultivariateNormal) -> None:
        r"""A posterior for a fully Bayesian model.

        Args:
            distribution: A GPyTorch MultivariateNormal (single-output case)
        """
        super().__init__(distribution=distribution)
        self._mean = (
            distribution.mean if self._is_mt else distribution.mean.unsqueeze(-1)
        )
        self._variance = (
            distribution.variance
            if self._is_mt
            else distribution.variance.unsqueeze(-1)
        )
        self._covariance_matrix = distribution.lazy_covariance_matrix

        self._mixture_mean: Tensor | None = None
        self._mixture_variance: Tensor | None = None
        self._mixture_covariance_matrix: Tensor | None = None

    @property
    def mixture_mean(self) -> Tensor:
        r"""The posterior mean for the mixture of models."""
        if self._mixture_mean is None:
            self._mixture_mean = self._mean.mean(dim=MCMC_DIM)
        return self._mixture_mean

    @property
    def mixture_variance(self) -> Tensor:
        r"""The posterior variance for the mixture of models."""
        if self._mixture_variance is None:
            num_mcmc_samples = self.mean.shape[MCMC_DIM]
            t1 = self._variance.sum(dim=MCMC_DIM) / num_mcmc_samples
            t2 = self._mean.pow(2).sum(dim=MCMC_DIM) / num_mcmc_samples
            t3 = -(self._mean.sum(dim=MCMC_DIM) / num_mcmc_samples).pow(2)
            self._mixture_variance = t1 + t2 + t3
        return self._mixture_variance

    @property
    def mixture_covariance_matrix(self) -> Tensor:
        r"""The posterior covariance matrix for the mixture of models."""
        if self._mixture_covariance_matrix is None:
            num_mcmc_samples = self.mean.shape[MCMC_DIM]
            t1 = self._covariance_matrix.sum(dim=MCMC_DIM) / num_mcmc_samples
            mean_diff = self._mean - self.mixture_mean.unsqueeze(MCMC_DIM)
            t2 = (
                torch.matmul(mean_diff, mean_diff.transpose(-1, -2)).sum(dim=MCMC_DIM)
                / num_mcmc_samples
            )
            self._mixture_covariance_matrix = t1 + t2

        return self._mixture_covariance_matrix

    def quantile(self, value: Tensor) -> Tensor:
        r"""Compute the posterior quantiles for the mixture of models."""
        return _quantile(posterior=self, value=value)

    @property
    def batch_range(self) -> tuple[int, int]:
        r"""The t-batch range.

        This is used in samplers to identify the t-batch component of the
        `base_sample_shape`. The base samples are expanded over the t-batches to
        provide consistency in the acquisition values, i.e., to ensure that a
        candidate produces same value regardless of its position on the t-batch.
        """
        return (0, -2) if self._is_mt else (0, -1)


class FullyBayesianPosterior(GaussianMixturePosterior):
    """For backwards compatibility."""

    def __init__(self, distribution: MultivariateNormal) -> None:
        """DEPRECATED."""
        warn(
            "`FullyBayesianPosterior` is marked for deprecation, consider using "
            "`GaussianMixturePosterior` instead.",
            DeprecationWarning,
        )
        super().__init__(distribution=distribution)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
A wrapper around AcquisitionFunctions to add proximal weighting of the
acquisition function.
"""

from __future__ import annotations

import torch
from botorch.acquisition import AcquisitionFunction
from botorch.exceptions.errors import UnsupportedError
from botorch.models import ModelListGP
from botorch.models.gpytorch import BatchedMultiOutputGPyTorchModel
from botorch.models.model import Model
from botorch.models.transforms.input import InputTransform
from botorch.utils import t_batch_mode_transform
from botorch.utils.transforms import average_over_ensemble_models
from torch import Tensor
from torch.nn import Module


class ProximalAcquisitionFunction(AcquisitionFunction):
    """A wrapper around AcquisitionFunctions to add proximal weighting of the
    acquisition function. The acquisition function is
    weighted via a squared exponential centered at the last training point,
    with varying lengthscales corresponding to `proximal_weights`. Can only be used
    with acquisition functions based on single batch models. Acquisition functions
    must be positive or `beta` must be specified to apply a SoftPlus transform before
    proximal weighting.

    Small values of `proximal_weights` corresponds to strong biasing towards recently
    observed points, which smoothes optimization with a small potential decrese in
    convergence rate.



    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> EI = ExpectedImprovement(model, best_f=0.0)
        >>> proximal_weights = torch.ones(d)
        >>> EI_proximal = ProximalAcquisitionFunction(EI, proximal_weights)
        >>> eip = EI_proximal(test_X)
    """

    def __init__(
        self,
        acq_function: AcquisitionFunction,
        proximal_weights: Tensor,
        transformed_weighting: bool | None = True,
        beta: float | None = None,
    ) -> None:
        r"""Derived Acquisition Function weighted by proximity to recently
        observed point.

        Args:
            acq_function: The base acquisition function, operating on input tensors
                of feature dimension `d`.
            proximal_weights: A `d` dim tensor used to bias locality
                along each axis.
            transformed_weighting: If True, the proximal weights are applied in
                the transformed input space given by
                `acq_function.model.input_transform` (if available), otherwise
                proximal weights are applied in real input space.
            beta: If not None, apply a softplus transform to the base acquisition
                function, allows negative base acquisition function values.
        """
        Module.__init__(self)

        self.acq_func = acq_function
        model = self.acq_func.model

        if hasattr(acq_function, "X_pending"):
            if acq_function.X_pending is not None:
                raise UnsupportedError(
                    "Proximal acquisition function requires `X_pending` to be None."
                )
            self.X_pending = acq_function.X_pending

        self.register_buffer("proximal_weights", proximal_weights)
        self.register_buffer(
            "transformed_weighting", torch.tensor(transformed_weighting)
        )

        self.register_buffer("beta", None if beta is None else torch.tensor(beta))

        _validate_model(model, proximal_weights)

    @t_batch_mode_transform(expected_q=1, assert_output_shape=False)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate base acquisition function with proximal weighting.

        Args:
            X: Input tensor of feature dimension `d` .

        Returns:
            Base acquisition function evaluated on tensor `X` multiplied by proximal
            weighting.
        """
        model = self.acq_func.model

        train_inputs = model.train_inputs[0]

        # if the model is ModelListGP then get the first model
        if isinstance(model, ModelListGP):
            train_inputs = train_inputs[0]
            model = model.models[0]

        # if the model has more than one output get the first copy of training inputs
        if isinstance(model, BatchedMultiOutputGPyTorchModel) and model.num_outputs > 1:
            train_inputs = train_inputs[0]

        input_transform = _get_input_transform(model)

        last_X = train_inputs[-1].reshape(1, 1, -1)

        # if transformed_weighting, transform X to calculate diff
        # (proximal weighting in transformed space)
        # otherwise,un-transform the last observed point to real space
        # (proximal weighting in real space)
        if input_transform is not None:
            if self.transformed_weighting:
                # transformed space weighting
                diff = input_transform.transform(X) - last_X
            else:
                # real space weighting
                diff = X - input_transform.untransform(last_X)

        else:
            # no transformation
            diff = X - last_X

        M = torch.linalg.norm(diff / self.proximal_weights, dim=-1) ** 2
        proximal_acq_weight = torch.exp(-0.5 * M)

        base_acqf = self.acq_func(X)
        if self.beta is None:
            if torch.any(base_acqf < 0):
                raise RuntimeError(
                    "Cannot use proximal biasing for negative "
                    "acquisition function values, set a value for beta to "
                    "fix this with a softplus transform"
                )

        else:
            base_acqf = torch.nn.functional.softplus(base_acqf, beta=self.beta)

        return base_acqf * proximal_acq_weight.flatten()


def _validate_model(model: Model, proximal_weights: Tensor) -> None:
    r"""Validate model

    Perform vaidation checks on model used in base acquisition function to make sure
    it is compatible with proximal weighting.

    Args:
        model: Model associated with base acquisition function to be validated.
        proximal_weights: A `d` dim tensor used to bias locality
                along each axis.
    """

    # check model for train_inputs and single batch
    if not hasattr(model, "train_inputs"):
        raise UnsupportedError("Acquisition function model must have `train_inputs`.")

    # get train inputs for each type of possible model
    if isinstance(model, ModelListGP):
        # ModelListGP models
        # check to make sure that the training inputs and input transformers for each
        # model match and are reversible
        train_inputs = model.train_inputs[0][0]
        input_transform = _get_input_transform(model.models[0])

        for i in range(len(model.train_inputs)):
            if not torch.equal(train_inputs, model.train_inputs[i][0]):
                raise UnsupportedError(
                    "Proximal acquisition function does not support unequal "
                    "training inputs"
                )

            if not input_transform == _get_input_transform(model.models[i]):
                raise UnsupportedError(
                    "Proximal acquisition function does not support non-identical "
                    "input transforms"
                )

    else:
        # any non-ModelListGP model
        train_inputs = model.train_inputs[0]

    # check to make sure that the model is single t-batch (q-batches are allowed)
    if model.batch_shape != torch.Size([]) and train_inputs.shape[1] != 1:
        raise UnsupportedError(
            "Proximal acquisition function requires a single batch model"
        )

    # check to make sure that weights match the training data shape
    if (
        len(proximal_weights.shape) != 1
        or proximal_weights.shape[0] != train_inputs.shape[-1]
    ):
        raise ValueError(
            "`proximal_weights` must be a one dimensional tensor with "
            "same feature dimension as model."
        )


def _get_input_transform(model: Model) -> InputTransform | None:
    """get input transform if defined"""
    try:
        return model.input_transform
    except AttributeError:
        return None



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
References

.. [lin2025scalable]
    J. A. Lin, S. Ament, M. Balandat, D. Eriksson, J. M. Hernández-Lobato, E. Bakshy.
    Scalable Gaussian Processes with Latent Kronecker Structure.
    International Conference on Machine Learning 2025.

.. [lin2024scaling]
    J. A. Lin, S. Ament, M. Balandat, E. Bakshy. Scaling Gaussian Processes
    for Learning Curve Prediction via Latent Kronecker Structure. NeurIPS 2024
    Bayesian Decision-making and Uncertainty Workshop.

.. [lin2023sampling]
    J. A. Lin, J. Antorán, s. Padhy, D. Janz, J. M. Hernández-Lobato, A. Terenin.
    Sampling from Gaussian Process Posterior using Stochastic Gradient Descent.
    Advances in Neural Information Processing Systems 2023.
"""

import contextlib
import warnings
from typing import Any

import torch
from botorch.acquisition.objective import PosteriorTransform
from botorch.exceptions.errors import BotorchTensorDimensionError
from botorch.exceptions.warnings import InputDataWarning
from botorch.models.gpytorch import GPyTorchModel
from botorch.models.model import FantasizeMixin, Model
from botorch.models.transforms.input import InputTransform
from botorch.models.transforms.outcome import OutcomeTransform, Standardize
from botorch.posteriors.gpytorch import GPyTorchPosterior
from botorch.posteriors.latent_kronecker import LatentKroneckerGPPosterior
from botorch.utils.datasets import SupervisedDataset
from botorch.utils.types import _DefaultType, DEFAULT
from gpytorch.distributions import MultivariateNormal
from gpytorch.kernels import MaternKernel, ScaleKernel
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.likelihoods.likelihood import Likelihood
from gpytorch.means import Mean, ZeroMean

from gpytorch.models.exact_gp import ExactGP
from gpytorch.module import Module
from linear_operator import settings
from linear_operator.operators import (
    ConstantDiagLinearOperator,
    KroneckerProductLinearOperator,
    MaskedLinearOperator,
)
from torch import Tensor


class LatentKroneckerGP(GPyTorchModel, ExactGP, FantasizeMixin):
    r"""
    A multi-task GP model which uses Kronecker structure despite missing entries.

    Leverages pathwise conditioning and iterative linear system solvers to
    efficiently draw samples from the GP posterior. See [lin2024scaling]_
    and [lin2025scalable]_ for details.

    For more information about pathwise conditioning, see [wilson2021pathwise]_
    and [Maddox2021bohdo]_. Details about iterative linear system solvers for GPs
    with pathwise conditioning can be found in [lin2023sampling]_.

    NOTE: This model requires iterative methods for efficient posterior inference.
    To enable iterative methods, the `use_iterative_methods` helper function can be
    used as a context manager.

    Example:
        >>> model = LatentKroneckerGP(train_X, train_T, train_Y)
        >>> mll = ExactMarginalLogLikelihood(model.likelihood, model)
        >>> with model.use_iterative_methods():
        >>>     fit_gpytorch_mll(mll)
        >>>     samples = model.posterior(test_X, test_T).rsample()
    """

    def __init__(
        self,
        train_X: Tensor,
        train_T: Tensor,
        train_Y: Tensor,
        likelihood: Likelihood | None = None,
        mean_module_X: Mean | None = None,
        mean_module_T: Mean | None = None,
        covar_module_X: Module | None = None,
        covar_module_T: Module | None = None,
        input_transform: InputTransform | None = None,
        outcome_transform: OutcomeTransform | _DefaultType | None = DEFAULT,
    ) -> None:
        r"""
        Args:
            train_X: A `batch_shape x n x d` tensor of training features.
            train_T: A `batch_shape x t x 1` tensor of training time steps.
            train_Y: A `batch_shape x n x t` tensor of training observations,
                corresponding to the Cartesian product of `train_X` and `train_T`.
            likelihood: A likelihood. If omitted, use a standard
                `GaussianLikelihood` with inferred homoskedastic noise level.
            mean_module_X: The mean function to be used for X.
                If omitted, a `ZeroMean` will be used.
            mean_module_T: The mean function to be used for T.
                If omitted, a `ZeroMean` will be used.
            covar_module_X: The module computing the covariance matrix of X.
                If omitted, a `MaternKernel` will be used.
            covar_module_T: The module computing the covariance matrix of T.
                If omitted, a `MaternKernel` wrapped in a `ScaleKernel` will be used.
            input_transform: An input transform that is applied to X.
            outcome_transform: An outcome transform that is applied to Y.
                Note that `.train()` will be called on the outcome transform during
                instantiation of the model.
            input_transform: An input transform that is applied in the model's
                forward pass.
        """
        with torch.no_grad():
            # transform inputs here to check resulting shapes
            # actual transforms will be applied in forward() and posterior()
            transformed_X = self.transform_inputs(
                X=train_X, input_transform=input_transform
            )

        self._validate_tensor_args(X=transformed_X, Y=train_Y)
        batch_shape, ard_num_dims = transformed_X.shape[:-2], transformed_X.shape[-1]
        self._num_outputs = train_Y.shape[-1]

        expected_shape = torch.Size([*batch_shape, self._num_outputs, 1])
        train_T = torch.broadcast_to(train_T, (*batch_shape, *train_T.shape[-2:]))
        if train_T.shape != expected_shape:
            raise BotorchTensorDimensionError(
                f"Expected train_T with shape {expected_shape} but got {train_T.shape}."
            )
        self.train_T = train_T

        mask_valid_batch = train_Y.isfinite()
        # flatten over batch_shape
        mask_valid_flat = mask_valid_batch.reshape(-1, *mask_valid_batch.shape[-2:])
        # check that all masks are equal across batch_shape
        if not torch.all((mask_valid_flat == mask_valid_flat[0]).all(dim=(-2, -1))):
            raise ValueError(
                "Pattern of missing values in train_Y must be equal across batch_shape."
            )

        self.mask_valid = mask_valid_flat[0].flatten()
        train_Y = train_Y.reshape(*batch_shape, -1)[..., self.mask_valid]

        if outcome_transform == DEFAULT:
            outcome_transform = Standardize(m=1, batch_shape=batch_shape)
        if outcome_transform is not None:
            outcome_transform.train()
            # transform outputs once and keep the results
            train_Y, _ = outcome_transform(train_Y.unsqueeze(-1), X=transformed_X)
            train_Y = train_Y.squeeze(-1)

        if likelihood is None:
            likelihood = GaussianLikelihood(batch_shape=batch_shape)

        ExactGP.__init__(
            self,
            train_inputs=train_X,
            train_targets=train_Y,
            likelihood=likelihood,
        )

        if mean_module_X is None:
            mean_module_X = ZeroMean(batch_shape=batch_shape)
        self.mean_module_X: Module = mean_module_X

        if mean_module_T is None:
            mean_module_T = ZeroMean(batch_shape=batch_shape)
        self.mean_module_T: Module = mean_module_T

        if covar_module_X is None:
            covar_module_X = MaternKernel(
                ard_num_dims=ard_num_dims, batch_shape=batch_shape
            )

        if covar_module_T is None:
            covar_module_T = ScaleKernel(
                base_kernel=MaternKernel(ard_num_dims=1, batch_shape=batch_shape),
            )

        self.covar_module_X: Module = covar_module_X
        self.covar_module_T: Module = covar_module_T

        if input_transform is not None:
            self.input_transform = input_transform
        if outcome_transform is not None:
            self.outcome_transform = outcome_transform

        self.to(train_X)

    def use_iterative_methods(
        self,
        tol: float = 0.01,
        max_iter: int = 10000,
        covar_root_decomposition: bool = False,
        log_prob: bool = True,
        solves: bool = True,
    ):
        with contextlib.ExitStack() as stack:
            stack.enter_context(
                settings.fast_computations(
                    covar_root_decomposition=covar_root_decomposition,
                    log_prob=log_prob,
                    solves=solves,
                )
            )
            stack.enter_context(settings.cg_tolerance(tol))
            stack.enter_context(settings.max_cg_iterations(max_iter))
            return stack.pop_all()

    def _get_mean(self, X: Tensor, T: Tensor, mask: Tensor | None = None) -> Tensor:
        mean_X = self.mean_module_X(X).unsqueeze(-1)
        mean_T = self.mean_module_T(T).unsqueeze(-1)
        mean = KroneckerProductLinearOperator(mean_X, mean_T).squeeze(-1)
        return mean[..., mask] if mask is not None else mean

    def forward(self, X: Tensor, T: Tensor | None = None) -> MultivariateNormal:
        r"""
        Computes the joint distribution at the given input locations.

        Args:
            X: A tensor of `X`-locations at which to compute the joint distribution.
            T: A tensor of `T`-locations at which to compute the joint distribution.
                If None, defaults to using `self.train_T`.

        Returns:
            MultivariateNormal: The joint distribution at the specified input locations.
        """
        if T is None:
            T = self.train_T

        if self.training:
            X = self.transform_inputs(X)
            mask = self.mask_valid
        else:
            num_outputs = X.shape[-2] * T.shape[-2]
            mask = torch.ones(num_outputs, dtype=torch.bool, device=X.device)
            mask[: self.mask_valid.shape[-1]] = self.mask_valid

        mean = self._get_mean(X, T, mask=mask)

        covar_X = self.covar_module_X(X)
        covar_T = self.covar_module_T(T)
        covar = KroneckerProductLinearOperator(covar_X, covar_T)
        covar = MaskedLinearOperator(covar, row_mask=mask, col_mask=mask)

        return MultivariateNormal(mean, covar)

    def posterior(
        self,
        X: Tensor,
        T: Tensor | None = None,
        observation_noise: bool | Tensor = False,
        posterior_transform: PosteriorTransform | None = None,
        **kwargs: Any,
    ) -> GPyTorchPosterior:
        r"""Computes the posterior over model outputs at the provided points.

        Args:
            X: A `(batch_shape) x q x d`-dim Tensor, where `d` is the dimension
                of the feature space and `q` is the number of points considered
                jointly.
            T: A `(batch_shape) x t x 1`-dim Tensor of `T`-locations at which to
                compute the posterior. If None, defaults to using `self.train_T`.
            observation_noise: If True, add the observation noise from the
                likelihood to the posterior. If a Tensor, use it directly as the
                observation noise (must be of shape `(batch_shape) x q`). It is
                assumed to be in the outcome-transformed space if an outcome
                transform is used.
            posterior_transform: An optional PosteriorTransform.

        Returns:
            A `GPyTorchPosterior` object, representing a batch of `b` joint
            distributions over `q` points. Includes observation noise if
            specified.
        """
        if posterior_transform is not None:
            raise NotImplementedError(
                "Posterior transforms currently not supported for "
                f"{self.__class__.__name__}"
            )
        if not isinstance(self.likelihood, GaussianLikelihood):
            raise NotImplementedError(
                "Only GaussianLikelihood currently supported for "
                f"{self.__class__.__name__}"
            )
        if observation_noise is not False:
            raise NotImplementedError(
                "Observation noise currently not supported for "
                f"{self.__class__.__name__}"
            )

        if T is None:
            T = self.train_T
        return LatentKroneckerGPPosterior(self, X, T)

    def _rsample_from_base_samples(
        self,
        X: Tensor,
        T: Tensor,
        base_samples: Tensor,
        observation_noise: bool | Tensor = False,
    ) -> Tensor:
        r"""Sample from the posterior distribution at the provided points `X`
        using Matheron's rule, requiring `n + 2 n_train` base samples.

        Args:
            X: A `(batch_shape) x q x d`-dim Tensor, where `d` is the dimension
                of the feature space and `q` is the number of points considered
                jointly
            T: A `(batch_shape) x t x 1`-dim Tensor of `T`-locations at which to
                evaluate the posterior samples.
            base_samples: A Tensor of `N(0, I)` base samples of shape
                `sample_shape x base_sample_shape`, typically obtained from
                a `Sampler`. This is used for deterministic optimization.
        Returns:
            Samples from the posterior, a tensor of shape
            `self._extended_shape(sample_shape=sample_shape)`.
        """
        # toggle eval mode to switch the behavior of input / outcome transforms
        # this also implicitly applies the input transform to the train_inputs
        self.eval()
        X_train = self.train_inputs[0]
        X_test = self.transform_inputs(X)
        n_train_full = X_train.shape[-2] * self._num_outputs
        n_train = self.train_targets.shape[-1]
        n_test = X_test.shape[-2] * T.shape[-2]

        sample_shape = base_samples.shape[: -len(self.batch_shape) - 1]
        w_train, eps_base, w_test = torch.split(
            base_samples, [n_train_full, n_train, n_test], dim=-1
        )
        eps = torch.sqrt(self.likelihood.noise) * eps_base

        # calculate prior sample evaluated at training data
        K_train_train_X = self.covar_module_X(X_train)
        K_train_train_T = self.covar_module_T(self.train_T)
        K_train_train = KroneckerProductLinearOperator(K_train_train_X, K_train_train_T)

        L_train_train_X = K_train_train_X.cholesky(upper=False)
        L_train_train_T = K_train_train_T.cholesky(upper=False)
        L_train_train = KroneckerProductLinearOperator(L_train_train_X, L_train_train_T)

        m_train = self._get_mean(X_train, self.train_T, mask=self.mask_valid)
        f_prior_train = L_train_train @ w_train.unsqueeze(-1)
        f_prior_train = m_train + f_prior_train.squeeze(-1)[..., self.mask_valid]

        # assemble and solve pathwise conditioning linear system
        K_train_train_valid = MaskedLinearOperator(
            K_train_train, row_mask=self.mask_valid, col_mask=self.mask_valid
        )
        noise_covar = ConstantDiagLinearOperator(
            self.likelihood.noise
            * torch.ones(*self.batch_shape, 1, dtype=X.dtype, device=X.device),
            diag_shape=n_train,
        )
        H = K_train_train_valid + noise_covar

        v = self.train_targets - (f_prior_train + eps)
        # expand once here to avoid repeated expansion
        # by MaskedLinearOperator later
        H_inv_v = torch.zeros(
            *sample_shape,
            *self.batch_shape,
            n_train_full,
            dtype=X.dtype,
            device=X.device,
        )
        with self.use_iterative_methods():
            H_inv_v[..., self.mask_valid] = H.solve(v.unsqueeze(-1)).squeeze(-1)

        # calculate prior sample evaluated at test data via conditional sampling
        K_test_test_X = self.covar_module_X(X_test).evaluate_kernel()
        K_test_test_T = self.covar_module_T(T).evaluate_kernel()
        K_train_test_X = self.covar_module_X(X_train, X_test).evaluate_kernel()
        K_train_test_T = self.covar_module_T(self.train_T, T).evaluate_kernel()

        L_train_test_X = L_train_train_X.solve_triangular(
            K_train_test_X.tensor, upper=False
        )
        L_train_test_T = L_train_train_T.solve_triangular(
            K_train_test_T.tensor, upper=False
        )

        L_test_test_X = (
            K_test_test_X - L_train_test_X.transpose(-2, -1) @ L_train_test_X
        ).cholesky(upper=False)
        L_test_test_T = (
            K_test_test_T - L_train_test_T.transpose(-2, -1) @ L_train_test_T
        ).cholesky(upper=False)

        L_test_train = KroneckerProductLinearOperator(
            L_train_test_X.transpose(-2, -1), L_train_test_T.transpose(-2, -1)
        )
        L_test_test = KroneckerProductLinearOperator(L_test_test_X, L_test_test_T)

        # match dimensions for broadcasting
        broadcast_shape = L_test_train.shape[:-2]
        extra_batch_dims = len(broadcast_shape) - len(self.batch_shape)
        for _ in range(extra_batch_dims):
            w_train = w_train.unsqueeze(len(sample_shape))
            w_test = w_test.unsqueeze(len(sample_shape))
            H_inv_v = H_inv_v.unsqueeze(len(sample_shape))

        m_test = self._get_mean(X_test, T)
        f_prior_test = L_test_train @ w_train.unsqueeze(-1)
        f_prior_test = f_prior_test + L_test_test @ w_test.unsqueeze(-1)
        f_prior_test = m_test + f_prior_test.squeeze(-1)

        K_train_test = KroneckerProductLinearOperator(K_train_test_X, K_train_test_T)
        # no MaskedLinearOperator here because H_inv_v is already expanded
        samples = K_train_test.transpose(-2, -1) @ H_inv_v.unsqueeze(-1)
        samples = samples + f_prior_test.unsqueeze(-1)
        # reshape samples to separate X and T dimensions
        # samples.shape = (*sample_shape, *broadcast_shape, n_test_x * n_t, 1)
        samples = samples.reshape(*samples.shape[:-2], X_test.shape[-2], T.shape[-2])
        # samples.shape = (*sample_shape, *broadcast_shape, n_test_x, n_t)
        if hasattr(self, "outcome_transform") and self.outcome_transform is not None:
            samples, _ = self.outcome_transform.untransform(samples, X=X)
        return samples

    def condition_on_observations(
        self, X: Tensor, Y: Tensor, noise: Tensor | None = None, **kwargs: Any
    ) -> Model:
        raise NotImplementedError(
            f"Conditioning currently not supported for {self.__class__.__name__}"
        )

    @classmethod
    def construct_inputs(cls, training_data: SupervisedDataset) -> dict[str, Any]:
        """
        Constructs the input tensors for LatentKroneckerGP from a SupervisedDataset.

        This method processes the provided training data to extract and organize the
        features and targets into the required format for the LatentKroneckerGP model.
        It factorizes inputs from the product space into the factors X and T.
        The matching output Y values are assembled by mapping observed values to their
        corresponding positions and filling missing values with NaN.

        Args:
            training_data: A SupervisedDataset containing training inputs and outputs.

        Returns:
            A dictionary with keys `train_X`, `train_T`, and `train_Y`, where:
                - `train_X`: The unique feature values (excluding the T dimension).
                - `train_T`: The unique feature values of the T dimension.
                - `train_Y`: The outputs aligned with the Cartesian product of
                    `train_X` and `train_T`, with missing values filled as NaN.
        """
        model_inputs = super().construct_inputs(training_data=training_data)

        if "train_Yvar" in model_inputs:
            warnings.warn(
                "Ignoring Yvar values in provided training data, because "
                "they are currently not supported by LatentKroneckerGP.",
                InputDataWarning,
                stacklevel=2,
            )

        t_idx = training_data.feature_names.index("step")
        x_idx = [i for i in range(len(training_data.feature_names)) if i != t_idx]

        # Factorize product space into factors X and T by finding unique values
        train_X, x_idx = model_inputs["train_X"][..., x_idx].unique(
            sorted=True, return_inverse=True, dim=-2
        )
        train_T, t_idx = model_inputs["train_X"][..., [t_idx]].unique(
            sorted=True, return_inverse=True, dim=-2
        )

        # Initialize train_Y with NaN for the full Cartesian product
        batch_shape = train_X.shape[:-2]
        n_x = train_X.shape[-2]
        n_t = train_T.shape[-2]
        train_Y = torch.full(
            (*batch_shape, n_x * n_t, 1),
            torch.nan,
            dtype=model_inputs["train_Y"].dtype,
            device=model_inputs["train_Y"].device,
        )

        # Convert 2D indices to 1D indices
        y_idx = x_idx * n_t + t_idx
        # Map original observations to their positions in the Cartesian product
        train_Y[..., y_idx, :] = model_inputs["train_Y"]
        train_Y = train_Y.reshape(*batch_shape, n_x, n_t)

        return {"train_X": train_X, "train_T": train_T, "train_Y": train_Y}



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
test_utils has its own directory with 'botorch/' to avoid circular dependencies:
Anything in 'tests/' can depend on anything in 'botorch/test_utils/', and
anything in 'botorch/test_utils/' can depend on anything in the rest of
'botorch/'.
"""

from botorch.test_utils.mock import mock_optimize

__all__ = ["mock_optimize"]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Deterministic Models: Simple wrappers that allow the usage of deterministic
mappings via the BoTorch Model and Posterior APIs.

Deterministic models are useful for expressing known input-output relationships
within the BoTorch Model API. This is useful e.g. for multi-objective
optimization with known objective functions (e.g. the number of parameters of a
Neural Network in the context of Neural Architecture Search is usually a known
function of the architecture configuration), or to encode cost functions for
cost-aware acquisition utilities. Cost-aware optimization is desirable when
evaluations have a cost that is heterogeneous, either in the inputs `X` or in a
particular fidelity parameter that directly encodes the fidelity of the
observation. `GenericDeterministicModel` supports arbitrary deterministic
functions, while `AffineFidelityCostModel` is a particular cost model for
multi-fidelity optimization. Other use cases of deterministic models include
representing approximate GP sample paths, e.g. Matheron paths obtained
with `get_matheron_path_model`, which allows them to be substituted in acquisition
functions or in other places where a `Model` is expected.
"""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable

import torch
from botorch.exceptions.errors import UnsupportedError
from botorch.models.ensemble import EnsembleModel
from botorch.models.model import Model, ModelList
from botorch.utils.transforms import is_ensemble
from torch import Size, Tensor


class DeterministicModel(EnsembleModel):
    """Abstract base class for deterministic models."""

    @abstractmethod
    def forward(self, X: Tensor) -> Tensor:
        r"""Compute the (deterministic) model output at X.

        Args:
            X: A `batch_shape x n x d`-dim input tensor `X`.

        Returns:
            A `batch_shape x n x m`-dimensional output tensor (the outcome
            dimension `m` must be explicit if `m=1`).
        """
        pass  # pragma: no cover

    def _forward(self, X: Tensor) -> Tensor:
        r"""Compatibilizes the `DeterministicModel` with `EnsemblePosterior`"""
        return self.forward(X=X).unsqueeze(-3)


class GenericDeterministicModel(DeterministicModel):
    r"""A generic deterministic model constructed from a callable.

    Example:
        >>> f = lambda x: x.sum(dim=-1, keep_dims=True)
        >>> model = GenericDeterministicModel(f)
    """

    def __init__(
        self,
        f: Callable[[Tensor], Tensor],
        num_outputs: int = 1,
        batch_shape: torch.Size | None = None,
    ) -> None:
        r"""
        Args:
            f: A callable mapping a `batch_shape x n x d`-dim input tensor `X`
                to a `batch_shape x n x m`-dimensional output tensor (the
                outcome dimension `m` must be explicit, even if `m=1`).
            num_outputs: The number of outputs `m`.
        """
        super().__init__()
        self._f = f
        self._num_outputs = num_outputs
        self._batch_shape = batch_shape

    @property
    def batch_shape(self) -> torch.Size | None:
        r"""The batch shape of the model."""
        return self._batch_shape

    def subset_output(self, idcs: list[int]) -> GenericDeterministicModel:
        r"""Subset the model along the output dimension.

        Args:
            idcs: The output indices to subset the model to.

        Returns:
            The current model, subset to the specified output indices.
        """

        def f_subset(X: Tensor) -> Tensor:
            return self._f(X)[..., idcs]

        return self.__class__(f=f_subset, num_outputs=len(idcs))

    def forward(self, X: Tensor) -> Tensor:
        r"""Compute the (deterministic) model output at X.

        Args:
            X: A `batch_shape x n x d`-dim input tensor `X`.

        Returns:
            A `batch_shape x n x m`-dimensional output tensor.
        """
        Y = self._f(X)
        batch_shape = Y.shape[:-2]
        # allowing for old behavior of not specifying the batch_shape
        if self.batch_shape is not None:
            try:
                torch.broadcast_shapes(self.batch_shape, batch_shape)
            except RuntimeError:
                raise ValueError(
                    "GenericDeterministicModel was initialized with batch_shape="
                    f"{self.batch_shape=} but the output of f has a batch_shape="
                    f"{batch_shape=} that is not broadcastable with it."
                )
        return Y


class AffineDeterministicModel(DeterministicModel):
    r"""An affine deterministic model."""

    def __init__(self, a: Tensor, b: Tensor | float = 0.01) -> None:
        r"""Affine deterministic model from weights and offset terms.

        A simple model of the form

            y[..., m] = b[m] + sum_{i=1}^d a[i, m] * X[..., i]

        Args:
            a: A `d x m`-dim tensor of linear weights, where `m` is the number
                of outputs (must be explicit if `m=1`)
            b: The affine (offset) term. Either a float (for single-output
                models or if the offset is shared), or a `m`-dim tensor (with
                different offset values for for the `m` different outputs).
        """
        if not a.ndim == 2:
            raise ValueError("a must be two-dimensional")
        if not torch.is_tensor(b):
            b = torch.tensor([b])
        if not b.ndim == 1:
            raise ValueError("b nust be one-dimensional")
        super().__init__()
        self.register_buffer("a", a)
        self.register_buffer("b", b.expand(a.size(-1)))
        self._num_outputs = a.size(-1)

    def subset_output(self, idcs: list[int]) -> AffineDeterministicModel:
        r"""Subset the model along the output dimension.

        Args:
            idcs: The output indices to subset the model to.

        Returns:
            The current model, subset to the specified output indices.
        """
        a_sub = self.a.detach()[..., idcs].clone()
        b_sub = self.b.detach()[..., idcs].clone()
        return self.__class__(a=a_sub, b=b_sub)

    def forward(self, X: Tensor) -> Tensor:
        return self.b + torch.einsum("...d,dm", X, self.a)


class PosteriorMeanModel(DeterministicModel):
    """A deterministic model that always returns the posterior mean."""

    def __init__(self, model: Model) -> None:
        r"""
        Args:
            model: The base model.
        """
        super().__init__()
        self.model = model

    def forward(self, X: Tensor) -> Tensor:
        return self.model.posterior(X).mean

    @property
    def num_outputs(self) -> int:
        r"""The number of outputs of the model."""
        return self.model.num_outputs

    @property
    def batch_shape(self) -> torch.Size:
        r"""The batch shape of the model."""
        return self.model.batch_shape


class FixedSingleSampleModel(DeterministicModel):
    r"""
    A deterministic model defined by a single sample `w`.

    Given a base model `f` and a fixed sample `w`, the model always outputs

        y = f_mean(x) + f_stddev(x) * w

    We assume the outcomes are uncorrelated here.
    """

    def __init__(
        self,
        model: Model,
        w: Tensor | None = None,
        dim: int | None = None,
        jitter: float | None = 1e-8,
        dtype: torch.dtype | None = None,
        device: torch.dtype | None = None,
    ) -> None:
        r"""
        Args:
            model: The base model.
            w: A 1-d tensor with length model.num_outputs.
                If None, draw it from a standard normal distribution.
            dim: dimensionality of w.
                If None and w is not provided, draw w samples of size model.num_outputs.
            jitter: jitter value to be added for numerical stability, 1e-8 by default.
            dtype: dtype for w if specified
            device: device for w if specified
        """
        super().__init__()
        self.model = model
        self._num_outputs = model.num_outputs
        self.jitter = jitter
        if w is None:
            self.w = (
                torch.randn(model.num_outputs, dtype=dtype, device=device)
                if dim is None
                else torch.randn(dim, dtype=dtype, device=device)
            )
        else:
            self.w = w

    def forward(self, X: Tensor) -> Tensor:
        post = self.model.posterior(X)
        return post.mean + torch.sqrt(post.variance + self.jitter) * self.w.to(X)


class MatheronPathModel(DeterministicModel):
    r"""A deterministic model that returns a Matheron path sample.

    A Matheron path is a continuous sample path of a GP, obtained by drawing
    random Fourier features from a GP prior and a pathwise update rule based on
    the observed data.

    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> matheron_model = MatheronPathModel(model)
        >>> output = matheron_model(test_X)
    """

    def __init__(
        self,
        model: Model,
        sample_shape: Size | None = None,
        ensemble_as_batch: bool = False,
        seed: int | None = None,
    ) -> None:
        r"""
        Args:
            model: The base model.
            sample_shape: The shape of the sample paths to be drawn, if an ensemble
                of sample paths is desired. If this is specified, the resulting
                deterministic model will behave as if the `sample_shape` is
                prepended to the model's `batch_shape`.
            ensemble_as_batch: If True, and model is an ensemble model, the resulting
                model will treat the ensemble dimension as a batch dimension.
            seed: Random seed for reproducible path generation. If None,
                no specific seed is set.
        """
        super().__init__()
        self.model = model

        # Validate model compatibility
        if isinstance(model, ModelList) and len(model.models) != model.num_outputs:
            raise UnsupportedError(
                "A model-list of multi-output models is not supported."
            )

        # Initialize path generation parameters
        self.sample_shape = Size() if sample_shape is None else sample_shape
        self.ensemble_as_batch = ensemble_as_batch

        # NOTE circular import in pathwise/utils.py otherwise
        from botorch.sampling.pathwise import draw_matheron_paths

        # Generate the Matheron path once during initialization
        if seed is not None:
            with torch.random.fork_rng():
                torch.manual_seed(seed)
                self._path = draw_matheron_paths(
                    model,
                    sample_shape=self.sample_shape,
                )
        else:
            self._path = draw_matheron_paths(model, sample_shape=self.sample_shape)
        self._path.set_ensemble_as_batch(ensemble_as_batch)
        self._is_ensemble = is_ensemble(model) or len(self.sample_shape) > 0

    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate the Matheron path at X.

        Args:
            X: A `batch_shape x n x d`-dim input tensor `X`.

        Returns:
            A `[sample_shape x] batch_shape x n x m`-dimensional output tensor.
        """
        if self.model.num_outputs == 1:
            # For single-output, add the output dimension
            return self._path(X).unsqueeze(-1)
        elif isinstance(self.model, ModelList):
            # For model list, stack the path outputs
            return torch.stack(self._path(X), dim=-1)
        else:
            # For multi-output models
            return self._path(X.unsqueeze(-3)).transpose(-1, -2)

    @property
    def num_outputs(self) -> int:
        r"""The number of outputs of the model."""
        return self.model.num_outputs

    @property
    def batch_shape(self) -> torch.Size:
        r"""The batch shape of the model."""
        return self.sample_shape + self.model.batch_shape



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from botorch.optim.closures.core import (
    ForwardBackwardClosure,
    NdarrayOptimizationClosure,
)
from botorch.optim.closures.model_closures import (
    get_loss_closure,
    get_loss_closure_with_grads,
)


__all__ = [
    "ForwardBackwardClosure",
    "get_loss_closure",
    "get_loss_closure_with_grads",
    "NdarrayOptimizationClosure",
]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import warnings

from typing import Callable

import numpy as np
import torch
from botorch.acquisition.multi_objective.objective import (
    IdentityMCMultiOutputObjective,
    MCMultiOutputObjective,
)
from botorch.acquisition.multioutput_acquisition import MultiOutputAcquisitionFunction
from botorch.exceptions import BotorchWarning
from botorch.utils.multi_objective.hypervolume import get_hypervolume_maximizing_subset
from botorch.utils.multi_objective.pareto import is_non_dominated
from torch import Tensor

try:
    from pymoo.algorithms.moo.nsga2 import NSGA2
    from pymoo.core.problem import Problem
    from pymoo.optimize import minimize
    from pymoo.termination.max_gen import MaximumGenerationTermination

    class BotorchPymooProblem(Problem):
        def __init__(
            self,
            n_var: int,
            n_obj: int,
            xl: np.ndarray,
            xu: np.ndarray,
            acqf: MultiOutputAcquisitionFunction,
            dtype: torch.dtype,
            device: torch.device,
            ref_point: Tensor | None = None,
            objective: MCMultiOutputObjective | None = None,
            constraints: list[Callable[[Tensor], Tensor]] | None = None,
        ) -> None:
            """PyMOO problem for optimizing the model posterior mean using NSGA-II.

            This is instantiated and used within `optimize_with_nsgaii` to define
            the optimization problem to interface with pymoo.

            This assumes maximization of all objectives.

            Args:
                n_var: The number of tunable parameters (`d`).
                n_obj: The number of objectives.
                xl: A `d`-dim np.ndarray of lower bounds for each tunable parameter.
                xu: A `d`-dim np.ndarray of upper bounds for each tunable parameter.
                acqf: A MultiOutputAcquisitionFunction.
                dtype: The torch dtype.
                device: The torch device.
                acqf: The acquisition function to optimize.
                ref_point: A list or tensor with `m` elements representing the reference
                    point (in the outcome space), which is treated as a lower bound
                    on the objectives, after applying `objective` to the samples.
                objective: The MCMultiOutputObjective under which the samples are
                    evaluated. Defaults to `IdentityMultiOutputObjective()`.
                    This can be used to determine which outputs of the
                    MultiOutputAcquisitionFunction should be used as
                    objectives/constraints in NSGA-II.
                constraints: A list of callables, each mapping a Tensor of dimension
                    `sample_shape x batch-shape x q x m` to a Tensor of dimension
                    `sample_shape x batch-shape x q`, where negative values imply
                    feasibility.
            """
            num_constraints = 0 if constraints is None else len(constraints)
            if ref_point is not None:
                num_constraints += ref_point.shape[0]
            super().__init__(
                n_var=n_var,
                n_obj=n_obj,
                n_ieq_constr=num_constraints,
                xl=xl,
                xu=xu,
                type_var=np.double,
            )
            self.botorch_acqf = acqf
            self.botorch_ref_point = ref_point
            self.botorch_objective = (
                IdentityMCMultiOutputObjective() if objective is None else objective
            )
            self.botorch_constraints = constraints
            self.torch_dtype = dtype
            self.torch_device = device

        def _evaluate(self, x: np.ndarray, out: dict[str, np.ndarray]) -> None:
            """Evaluate x with respect to the objective/constraints."""
            X = torch.from_numpy(x).to(dtype=self.torch_dtype, device=self.torch_device)
            with torch.no_grad():
                # eval in batch mode, since all we need is the mean and this helps
                # avoid ill-conditioning
                y = self.botorch_acqf(X=X.unsqueeze(-2))
            obj = self.botorch_objective(y)
            # negate the objectives, since we want to maximize this function
            out["F"] = -obj.cpu().numpy()
            constraint_vals = None
            if self.botorch_constraints is not None:
                constraint_vals = torch.stack(
                    [c(y) for c in self.botorch_constraints], dim=-1
                )
            if self.botorch_ref_point is not None:
                # add constraints for the ref point
                ref_constraints = self.botorch_ref_point - obj
                if constraint_vals is not None:
                    constraint_vals = torch.cat(
                        [constraint_vals, ref_constraints], dim=-1
                    )
                else:
                    constraint_vals = ref_constraints
            if constraint_vals is not None:
                out["G"] = constraint_vals.cpu().numpy()

    def optimize_with_nsgaii(
        acq_function: MultiOutputAcquisitionFunction,
        bounds: Tensor,
        num_objectives: int,
        q: int | None = None,
        ref_point: list[float] | Tensor | None = None,
        objective: MCMultiOutputObjective | None = None,
        constraints: list[Callable[[Tensor], Tensor]] | None = None,
        population_size: int = 250,
        max_gen: int | None = None,
        seed: int | None = None,
        fixed_features: dict[int, float] | None = None,
    ) -> tuple[Tensor, Tensor]:
        """Optimize the posterior mean via NSGA-II, returning the Pareto set and front.

        This assumes maximization of all objectives.

        TODO: Add support for discrete parameters.

        Args:
            acq_function: The MultiOutputAcquisitionFunction to optimize.
            bounds: A `2 x d` tensor of lower and upper bounds for each column of `X`.
            q: The number of candidates. If None, return the full population.
            num_objectives: The number of objectives.
            ref_point: A list or tensor with `m` elements representing the reference
                point (in the outcome space), which is treated as a lower bound
                on the objectives, after applying `objective` to the samples.
            objective: The MCMultiOutputObjective under which the samples are
                evaluated. Defaults to `IdentityMultiOutputObjective()`.
                This can be used to determine which outputs of the
                MultiOutputAcquisitionFunction should be used as
                objectives/constraints in NSGA-II.
            constraints: A list of callables, each mapping a Tensor of dimension
                `sample_shape x batch-shape x q x m` to a Tensor of dimension
                `sample_shape x batch-shape x q`, where negative values imply
                feasibility.
            population_size: the population size for NSGA-II.
            max_gen: The number of iterations for NSGA-II. If None, this uses the
                default termination condition in pymoo for NSGA-II.
            seed: The random seed for NSGA-II.
            fixed_features: A map `{feature_index: value}` for features that
                should be fixed to a particular value during generation. All indices
                should be non-negative.

        Returns:
            A two-element tuple containing the pareto set X and pareto frontier Y.
        """
        tkwargs = {"dtype": bounds.dtype, "device": bounds.device}
        if ref_point is not None:
            ref_point = torch.as_tensor(ref_point, **tkwargs)
        if fixed_features is not None:
            bounds = bounds.clone()
            # set lower and upper bounds to the fixed value
            for i, val in fixed_features.items():
                bounds[:, i] = val
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=DeprecationWarning)
            pymoo_problem = BotorchPymooProblem(
                n_var=bounds.shape[-1],
                n_obj=num_objectives,
                xl=bounds[0].cpu().numpy(),
                xu=bounds[1].cpu().numpy(),
                acqf=acq_function,
                ref_point=ref_point,
                objective=objective,
                constraints=constraints,
                **tkwargs,
            )
            if q is not None:
                population_size = max(population_size, q)
            algorithm = NSGA2(pop_size=population_size, eliminate_duplicates=True)
            res = minimize(
                problem=pymoo_problem,
                algorithm=algorithm,
                termination=(
                    None
                    if max_gen is None
                    else MaximumGenerationTermination(n_max_gen=max_gen)
                ),
                seed=seed,
                verbose=False,
            )
        X = torch.tensor(res.X, **tkwargs)
        # multiply by negative one to return the correct sign for maximization
        Y = -torch.tensor(res.F, **tkwargs)
        pareto_mask = is_non_dominated(Y, deduplicate=True)
        X_pareto = X[pareto_mask]
        Y_pareto = Y[pareto_mask]
        if q is not None:
            if Y_pareto.shape[0] > q:
                Y_pareto, indices = get_hypervolume_maximizing_subset(
                    # use nadir as reference point since we likely don't care about the
                    # extrema as much as the interior
                    n=q,
                    Y=Y_pareto,
                    ref_point=Y_pareto.min(dim=0).values,
                )
                X_pareto = X_pareto[indices]
            elif Y_pareto.shape[0] < q:
                n_missing = q - Y_pareto.shape[0]
                if Y.shape[0] >= q:
                    # select some dominated solutions
                    rand_idcs = np.random.choice(
                        (~pareto_mask).nonzero().view(-1).cpu().numpy(),
                        n_missing,
                        replace=False,
                    )
                    rand_idcs = torch.from_numpy(rand_idcs).to(
                        device=pareto_mask.device
                    )
                    pareto_mask[rand_idcs] = 1
                    X_pareto = X[pareto_mask]
                    Y_pareto = Y[pareto_mask]
                else:
                    warnings.warn(
                        f"NSGA-II only returned {Y.shape[0]} points.",
                        BotorchWarning,
                        stacklevel=3,
                    )
                    return X, Y
        return X_pareto, Y_pareto

except ImportError:  # pragma: no cover
    pass
